VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
END
Attribute VB_Name = "clsDFA"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

'
' Class clsDFA (clsDFA.cls)
'
' Purpose:
'   Simulates a Deterministic Finite Automaton (DFA).
'
'   Simulates a DFA by running it step-by-step (the class raises an event every time
'   a new state has been reached). It can also save and load DFAs and run on existing
'   tapes (loaded from files or created randomly).
'


'
' Public Types
'


' Used for indicating the status of the DFA simulation.
Public Enum SimulationStatus
    ssStopped
    ssRunning
    ssPaused
    ssComplete
End Enum


' Used for indicating errors raised by the DFA.
Public Enum DFAError
    deInvalidIndex
    dePropertyCannotBeAccessedWhileSimulatorRunning
    deSubCannotRunWhileSimulatorRunning
    deInvalidStateID
    deInvalidState
    deNoDFAStates
    deNoInitialDFAState
    deNoCurrentDFAState
    deInvalidReadHeadPosition
    deDFAStateNotFound
    deInvalidStateDirection
    deStateAlreadyExists
    deFileDoesntExist
    deInvalidDFAFile
    deInvalidTapeFile
    deInvalidCellValue
    deInvalidDFAProperty
    deInvalidDFAStatement
    deInvalidDFAStateValue
    deInvalidTapeProperty
    deInvalidTapeStatement
End Enum


'
' Private Variables
'


' The default no. of cells in the tape read by the DFA read head (since the
' computer's memory is limited the tape must be finite).
Private Const DEFAULT_TAPE_SIZE As Integer = 1000
' The default tape cell value.
Private Const DEFAULT_TAPE_CELL_VALUE As String = "0"

' General information about the DFA (Name, description and author).
Private mstrDFAName As String
Private mstrDFADescription As String
Private mstrDFAAuthor As String
' The collection of states found in the DFA - each item in the collection contains a
' single DFA state, and has the following key: "StateID:CurrentCellContents".
Private mcolDFAStates As Collection
' The values of the tape cells (the default size of the tape is defined by the constant DEFAULT_TAPE_SIZE).
Private mstrTapeCells() As String
' The current position of the read head within the tape.
Private mintCurrentTapePosition As Integer
' Indicates which tape cells are breakpoints.
Private mblnTapeCellsBreakpoints() As Boolean
' The state ID we're currently at.
Private mstrCurrentStateID As String
' The initial state ID (the state where we start the simulation).
Private mstrInitialStateID As String
' Used for indicating the status of the DFA simulation.
Private mssSimulationStatus As SimulationStatus

' Advanced Options

' If set to true, the cells within the tape can contain values other than 0,1.
Private mblnTextualTapeCellValues As Boolean
' If set to true, the read head can be instructed by the DFA states to move more
' than one step to its left\right.
Private mblnReadHeadCanMoveSeveralSteps As Boolean
' If set to true, the read head can be instructed by the DFA states to stay at its
' current position and not move left\right.
Private mblnReadHeadCanStayAtPosition As Boolean
' If set to true, the DFA will stop running when reaching accepting states.
Private mblnAllowAcceptingStates As Boolean


'
' Public Events
'


' Raised before reaching a certain state.
'   CurrentState: The state the DFA has reached.
'   Pause: If set to True by the user during execution of the event, the DFA will
'          pause (and it will only continue if Pause is set to True).
'          Note: In order for Pause to be passed by *real* reference (so it can be changed
'                not only by the event sub, but by all subs), an object must be used
'                to hold the boolean value - in this case, a collection was selected.
Public Event ReachedState(ByRef CurrentState As clsDFAState, ByRef Pause As Collection)
' Raised when the simulation stops.
'   LatestState: The last state executed before the simulation stopped.
Public Event SimulationComplete(ByVal LatestState As clsDFAState)
' Raised when the simulation reaches a breakpoint.
'   BreakpointType : If 0, rule breakpoint; if 1, tape cell breakpoint.
Public Event ReachedBreakpoint(BreakpointType As Integer)
' Raised when an error is encountered during the execution of the DFA.
Public Event ErrorRaised(Error As DFAError, Description As String, Source As String)


'
' Public Properties
'


'
' DFAName Property
'

Public Property Get DFAName() As String
    DFAName = mstrDFAName
End Property

Public Property Let DFAName(NewName As String)
    mstrDFAName = NewName
End Property


'
' DFADescription Property
'

Public Property Get DFADescription() As String
    DFADescription = mstrDFADescription
End Property

Public Property Let DFADescription(NewDescription As String)
    mstrDFADescription = NewDescription
End Property


'
' DFAAuthor Property
'

Public Property Get DFAAuthor() As String
    DFAAuthor = mstrDFAAuthor
End Property

Public Property Let DFAAuthor(NewAuthor As String)
    mstrDFAAuthor = NewAuthor
End Property


'
' SimulationStatus Property (Read Only)
'

Public Property Get SimulationStatus() As SimulationStatus
    SimulationStatus = mssSimulationStatus
End Property

Public Property Let SimulationStatus(NewStatus As SimulationStatus)

    If (NewStatus = ssStopped) Then
        ' Stop (reset) the DFA simulation.
        Call ResetSimulation
    ElseIf (NewStatus = ssPaused) Then
        mssSimulationStatus = ssPaused
    ElseIf (NewStatus = ssRunning) Then
        ' Run the DFA simulation.
        Call RunSimulation
    End If
End Property


'
' DFAStates Property
'

Public Property Get DFAStates(Index As Integer) As clsDFAState
    On Error Resume Next
    
    Set DFAStates = mcolDFAStates(Index)
    
    ' Invalid index.
    If (DFAStates Is Nothing) Then _
        RaiseEvent ErrorRaised(deInvalidIndex, "Invalid index entered: " & Index, "clsDFA.Property Get DFAStates"): Exit Property
End Property

Public Property Let DFAStates(Index As Integer, NewState As clsDFAState)
    Dim clsTemp As clsDFAState
    
    On Error Resume Next
    
    Set clsTemp = mcolDFAStates(Index)
    
    ' Invalid index.
    If (clsTemp Is Nothing) Then _
        RaiseEvent ErrorRaised(deInvalidIndex, "Invalid index entered: " & Index, "clsDFA.Property Let DFAStates"): Exit Property
    
    ' The simulation mustn't be running in order to change this property.
    If (mssSimulationStatus = ssRunning) Then _
        RaiseEvent ErrorRaised(dePropertyCannotBeAccessedWhileSimulatorRunning, "The property cannot be accessed while the DFA simulator is running", "clsDFA.Property Let DFAStates"): Exit Property
    
    Set mcolDFAStates(Index) = NewState
    
    RefreshStatesCollection
End Property


'
' TapeCells Property
'

Public Property Get TapeCells(Index As Integer) As String
    ' Invalid position.
    If ((Index < LBound(mstrTapeCells)) Or _
        (Index > UBound(mstrTapeCells))) Then _
        RaiseEvent ErrorRaised(deInvalidIndex, "Invalid position entered: " & Index, "clsDFA.Property Get TapeCells"): Exit Property
    
    TapeCells = mstrTapeCells(Index)
End Property

Public Property Let TapeCells(Index As Integer, NewValue As String)
    ' Invalid position.
    If ((Index < LBound(mstrTapeCells)) Or _
        (Index > UBound(mstrTapeCells))) Then _
        RaiseEvent ErrorRaised(deInvalidIndex, "Invalid position entered: " & Index, "clsDFA.Property Let TapeCells"): Exit Property
    
    ' The simulation mustn't be running in order to change this property.
    If (mssSimulationStatus = ssRunning) Then _
        RaiseEvent ErrorRaised(dePropertyCannotBeAccessedWhileSimulatorRunning, "The property cannot be accessed while the DFA simulator is running", "clsDFA.Property Let TapeCells"): Exit Property
    
    mstrTapeCells(Index) = NewValue
End Property


'
' TapeCellsBreakPoint Property
'

Public Property Get TapeCellsBreakPoint(Index As Integer) As Boolean
    ' Invalid position.
    If ((Index < LBound(mblnTapeCellsBreakpoints)) Or _
        (Index > UBound(mblnTapeCellsBreakpoints))) Then _
        RaiseEvent ErrorRaised(deInvalidIndex, "Invalid position entered: " & Index, "clsDFA.Property Get TapeCellsBreakPoint"): Exit Property
    
    TapeCellsBreakPoint = mblnTapeCellsBreakpoints(Index)
End Property

Public Property Let TapeCellsBreakPoint(Index As Integer, NewValue As Boolean)
    ' Invalid position.
    If ((Index < LBound(mblnTapeCellsBreakpoints)) Or _
        (Index > UBound(mblnTapeCellsBreakpoints))) Then _
        RaiseEvent ErrorRaised(deInvalidIndex, "Invalid position entered: " & Index, "clsDFA.Property Let TapeCells"): Exit Property
    
    ' The simulation mustn't be running in order to change this property.
    If (mssSimulationStatus = ssRunning) Then _
        RaiseEvent ErrorRaised(dePropertyCannotBeAccessedWhileSimulatorRunning, "The property cannot be accessed while the DFA simulator is running", "clsDFA.Property Let TapeCellsBreakPoint"): Exit Property
    
    mblnTapeCellsBreakpoints(Index) = NewValue
End Property


'
' CurrentTapePosition Property
'

Public Property Get CurrentTapePosition() As Integer
    CurrentTapePosition = mintCurrentTapePosition
End Property

Public Property Let CurrentTapePosition(NewPosition As Integer)
    ' Invalid position.
    If ((NewPosition < LBound(mstrTapeCells)) Or _
        (NewPosition > UBound(mstrTapeCells))) Then _
        RaiseEvent ErrorRaised(deInvalidIndex, "Invalid position entered: " & NewPosition, "clsDFA.Property Let CurrentTapePosition"): Exit Property
    
    ' The simulation mustn't be running in order to change this property.
    If (mssSimulationStatus = ssRunning) Then _
        RaiseEvent ErrorRaised(dePropertyCannotBeAccessedWhileSimulatorRunning, "The property cannot be accessed while the DFA simulator is running", "clsDFA.Property Let CurrentTapePosition"): Exit Property
    
    mintCurrentTapePosition = NewPosition
End Property


'
' TapeCellCount Property
'

Public Property Get TapeCellCount() As Integer
    TapeCellCount = UBound(mstrTapeCells) - LBound(mstrTapeCells) + 1
End Property

Public Property Let TapeCellCount(NewSize As Integer)
    Dim intOldUBound As Integer
    Dim i As Integer
    
    ' Invalid position.
    If (NewSize < LBound(mstrTapeCells)) Then _
        RaiseEvent ErrorRaised(deInvalidIndex, "Invalid size entered: " & NewSize, "clsDFA.Property Let TapeCellCount"): Exit Property
    
    ' The simulation mustn't be running in order to change this property.
    If (mssSimulationStatus = ssRunning) Then _
        RaiseEvent ErrorRaised(dePropertyCannotBeAccessedWhileSimulatorRunning, "The property cannot be accessed while the DFA simulator is running", "clsDFA.Property Let TapeCellCount"): Exit Property
    
    intOldUBound = UBound(mstrTapeCells)
    
    ReDim Preserve mstrTapeCells(LBound(mstrTapeCells) To NewSize)
    ReDim Preserve mblnTapeCellsBreakpoints(LBound(mblnTapeCellsBreakpoints) To NewSize)
    
    If (NewSize > intOldUBound) Then
        ' Fill all new cells with a default value.
        For i = intOldUBound + 1 To NewSize
            mstrTapeCells(i) = DEFAULT_TAPE_CELL_VALUE
        Next i
    End If
    
    ' Change the current tape position if it's bigger than the new tape size.
    If (mintCurrentTapePosition > NewSize) Then _
        mintCurrentTapePosition = NewSize
End Property


'
' StateCount Property (Read Only)
'

Public Property Get StateCount() As Integer
    StateCount = mcolDFAStates.Count
End Property


'
' CurrentStateID Property
'

Public Property Get CurrentStateID() As String
    CurrentStateID = mstrCurrentStateID
End Property

Public Property Let CurrentStateID(NewStateID As String)
    ' Invalid state ID.
    If (Not StateIDExists(NewStateID)) Then _
        RaiseEvent ErrorRaised(deInvalidStateID, "Invalid DFA State ID entered: " & NewStateID, "clsDFA.Property Let CurrentStateID"): Exit Property
    ' The simulation mustn't be running in order to change this property.
    If (mssSimulationStatus = ssRunning) Then _
        RaiseEvent ErrorRaised(dePropertyCannotBeAccessedWhileSimulatorRunning, "The property cannot be accessed while the DFA simulator is running", "clsDFA.Property Let CurrentStateID"): Exit Property
    
    mstrCurrentStateID = NewStateID
End Property

'
' InitialState Property
'

Public Property Get InitialStateID() As String
    InitialStateID = mstrInitialStateID
End Property

Public Property Let InitialStateID(NewStateID As String)
    ' Invalid state ID.
    If (Not StateIDExists(NewStateID)) Then _
        RaiseEvent ErrorRaised(deInvalidStateID, "Invalid DFA State ID entered: " & NewStateID, "clsDFA.Property Let InitialStateID"): Exit Property
    ' The simulation mustn't be running in order to change this property.
    If (mssSimulationStatus = ssRunning) Then _
        RaiseEvent ErrorRaised(dePropertyCannotBeAccessedWhileSimulatorRunning, "The property cannot be accessed while the DFA simulator is running", "clsDFA.Property Let InitialStateID"): Exit Property
    
    mstrInitialStateID = NewStateID
End Property


'
' TextualTapeCellValues Property
'

Public Property Get TextualTapeCellValues() As Boolean
    TextualTapeCellValues = mblnTextualTapeCellValues
End Property

Public Property Let TextualTapeCellValues(Textual As Boolean)
    Dim i As Integer, t As Integer
    
    ' The simulation mustn't be running in order to change this property.
    If (mssSimulationStatus = ssRunning) Then _
        RaiseEvent ErrorRaised(dePropertyCannotBeAccessedWhileSimulatorRunning, "The property cannot be accessed while the DFA simulator is running", "clsDFA.Property Let TextualTapeCellValues"): Exit Property
    
    mblnTextualTapeCellValues = Textual
    
    If (Textual = False) Then
        ' Convert all non-zero\non-one cell values to zero.
        For i = LBound(mstrTapeCells) To UBound(mstrTapeCells)
            ' Convert the current tape cell to zero since it's not zero\one.
            If ((mstrTapeCells(i) <> "0") And (mstrTapeCells(i) <> "1")) Then _
                mstrTapeCells(i) = "0"
        Next i
    End If
End Property


'
' ReadHeadCanMoveSeveralSteps Property
'

Public Property Get ReadHeadCanMoveSeveralSteps() As Boolean
    ReadHeadCanMoveSeveralSteps = mblnReadHeadCanMoveSeveralSteps
End Property

Public Property Let ReadHeadCanMoveSeveralSteps(SeveralSteps As Boolean)
    Dim clsCurrentState As clsDFAState
    
    ' The simulation mustn't be running in order to change this property.
    If (mssSimulationStatus = ssRunning) Then _
        RaiseEvent ErrorRaised(dePropertyCannotBeAccessedWhileSimulatorRunning, "The property cannot be accessed while the DFA simulator is running", "clsDFA.Property Let ReadHeadCanMoveSeveralSteps"): Exit Property
    
    If ((mblnReadHeadCanMoveSeveralSteps = True) And (SeveralSteps = False)) Then
        ' Change any states which move the read head several steps to only one step left\right.
        
        For Each clsCurrentState In mcolDFAStates
            If (clsCurrentState.NewDirection > 1) Then
                clsCurrentState.NewDirection = 1
            ElseIf (clsCurrentState.NewDirection < -1) Then
                clsCurrentState.NewDirection = -1
            End If
        Next clsCurrentState
    End If
    
    mblnReadHeadCanMoveSeveralSteps = SeveralSteps
End Property


'
' ReadHeadCanStayAtPosition Property
'

Public Property Get ReadHeadCanStayAtPosition() As Boolean
    ReadHeadCanStayAtPosition = mblnReadHeadCanStayAtPosition
End Property

Public Property Let ReadHeadCanStayAtPosition(Stay As Boolean)
    Dim clsCurrentState As clsDFAState
    
    ' The simulation mustn't be running in order to change this property.
    If (mssSimulationStatus = ssRunning) Then _
        RaiseEvent ErrorRaised(dePropertyCannotBeAccessedWhileSimulatorRunning, "The property cannot be accessed while the DFA simulator is running", "clsDFA.Property Let ReadHeadCanStayAtPosition"): Exit Property
    
    If ((mblnReadHeadCanStayAtPosition = True) And (Stay = False)) Then
        ' Change any states which make the read head stay at the same position to move left\right.
        
        For Each clsCurrentState In mcolDFAStates
            If (clsCurrentState.NewDirection = 0) Then _
                clsCurrentState.NewDirection = 1
        Next clsCurrentState
    End If
    
    mblnReadHeadCanStayAtPosition = Stay
End Property


'
' AllowAcceptingStates Property
'

Public Property Get AllowAcceptingStates() As Boolean
    AllowAcceptingStates = mblnAllowAcceptingStates
End Property

Public Property Let AllowAcceptingStates(Accept As Boolean)
    Dim clsCurrentState As clsDFAState
    
    ' The simulation mustn't be running in order to change this property.
    If (mssSimulationStatus = ssRunning) Then _
        RaiseEvent ErrorRaised(dePropertyCannotBeAccessedWhileSimulatorRunning, "The property cannot be accessed while the DFA simulator is running", "clsDFA.Property Let AllowAcceptingStates"): Exit Property
    
    If ((mblnAllowAcceptingStates = True) And (Accept = False)) Then
        ' Change any accepting states to non-accepting states.
        
        For Each clsCurrentState In mcolDFAStates
            If (clsCurrentState.AcceptingState = True) Then _
                clsCurrentState.AcceptingState = False
        Next clsCurrentState
    End If
    
    mblnAllowAcceptingStates = Accept
End Property


'
' Public Functions
'


'
' Function ValidateStates
'
' Purpose:
'   Checks if all of the states added are valid.
'
' Parameters:
'   RaiseErrors (Optional) : If true, any time an invalid state is found, and error
'                            will be raised, specifying what's wrong with the state.
'
' Returns:
'   True if all states are valid, False otherwise.
'
Public Function ValidateStates(Optional RaiseErrors As Boolean = True) As Boolean
    Dim clsCurrentState As clsDFAState
    
    ' Initial state ID doesn't exist.
    If (StateIDExists(mstrInitialStateID) = False) Then
        RaiseEvent ErrorRaised(deNoInitialDFAState, "The initial state ID doesn' exist!", "clsDFA.ValidateStates")
        ValidateStates = False
        Exit Function
    End If
    
    For Each clsCurrentState In mcolDFAStates
        ' Check the current state.
        
        If (StateIDExists(clsCurrentState.NewStateID) = False) Then
            ' New State ID doesn't exist.
            If (RaiseErrors = True) Then _
                RaiseEvent ErrorRaised(deInvalidStateID, "The New State ID, '" & clsCurrentState.NewStateID & "', of the state '" & clsCurrentState.StateID & "' with the current cell contents of '" & clsCurrentState.CurrentCellContents & "', doesn't exist!", "clsDFA.ValidateStates")
            ValidateStates = False
            Exit Function
        ElseIf ((clsCurrentState.CurrentCellContents <> "0") And _
            (clsCurrentState.CurrentCellContents <> "1") And _
            (mblnTextualTapeCellValues = False)) Then
            ' Invalid Current Cell Contents.
            If (RaiseErrors = True) Then _
                RaiseEvent ErrorRaised(deInvalidCellValue, "The Current Cell Contents, '" & clsCurrentState.CurrentCellContents & "', of the state '" & clsCurrentState.StateID & "' with the current cell contents of '" & clsCurrentState.CurrentCellContents & "', is invalid (can only be either 0 or 1)!", "clsDFA.ValidateStates")
            ValidateStates = False
            Exit Function
        ElseIf ((clsCurrentState.NewCellContents <> "0") And _
            (clsCurrentState.NewCellContents <> "1") And _
            (mblnTextualTapeCellValues = False)) Then
            ' Invalid New Cell Contents.
            If (RaiseErrors = True) Then _
                RaiseEvent ErrorRaised(deInvalidCellValue, "The New Cell Contents, '" & clsCurrentState.NewCellContents & "', of the state '" & clsCurrentState.StateID & "' with the current cell contents of '" & clsCurrentState.CurrentCellContents & "', is invalid (can only be either 0 or 1)!", "clsDFA.ValidateStates")
            ValidateStates = False
            Exit Function
        ElseIf ((clsCurrentState.NewDirection = 0) And _
            (mblnReadHeadCanStayAtPosition = False)) Then
            ' Invalid New Direction.
            If (RaiseErrors = True) Then _
                RaiseEvent ErrorRaised(deInvalidCellValue, "The New Direction (stay at the same position) of the state '" & clsCurrentState.StateID & "' with the current cell contents of '" & clsCurrentState.CurrentCellContents & "', is invalid (the read head cannot stay at the same position)!", "clsDFA.ValidateStates")
            ValidateStates = False
            Exit Function
        ElseIf ((Abs(clsCurrentState.NewDirection) <> 1) And _
            (mblnReadHeadCanMoveSeveralSteps = False)) Then
            ' Invalid New Direction.
            If (RaiseErrors = True) Then _
                RaiseEvent ErrorRaised(deInvalidCellValue, "The New Direction (" & Abs(clsCurrentState.NewDirection) & IIf(clsCurrentState.NewDirection > 1, " Steps Right", " Steps Left") & ") of the state '" & clsCurrentState.StateID & "' with the current cell contents of '" & clsCurrentState.CurrentCellContents & "', is invalid (the read head cannot move than one step left\right)!", "clsDFA.ValidateStates")
            ValidateStates = False
            Exit Function
        ElseIf ((clsCurrentState.AcceptingState = True) And _
            (mblnAllowAcceptingStates = False)) Then
            ' Invalid Accepting State.
            If (RaiseErrors = True) Then _
                RaiseEvent ErrorRaised(deInvalidCellValue, "The current state, '" & clsCurrentState.StateID & "', with the current cell contents of '" & clsCurrentState.CurrentCellContents & "', cannot be an accepting state!", "clsDFA.ValidateStates")
            ValidateStates = False
            Exit Function
        End If
    Next clsCurrentState
    
    ValidateStates = True
End Function


'
' Function StateIDExists
'
' Purpose:
'   Checks if a certain DFA state ID exists.
'
' Parameters:
'   StateID : The ID of the state.
'
' Returns:
'   True if the DFA state ID exists, False otherwise.
'
Public Function StateIDExists(StateID As String) As Boolean
    Dim clsCurrentState As clsDFAState
    
    For Each clsCurrentState In mcolDFAStates
        If (clsCurrentState.StateID = StateID) Then
            ' Found a matching state ID.
            StateIDExists = True
            Exit Function
        End If
    Next clsCurrentState
    
    ' If we've reached this far, this means we didn't find a matching state ID.
    StateIDExists = False
End Function


'
' Function DFAStateExists
'
' Purpose:
'   Checks if a certain DFA state exists.
'
' Parameters:
'   StateID : The ID of the state.
'   CurrentCellContents : The current cell value of the state.
'
' Returns:
'   True if the DFA state exists, False otherwise.
'
Public Function DFAStateExists(StateID As String, CurrentCellContents As String) As Boolean
    Dim clsTemp As clsDFAState
    
    Set clsTemp = GetDFAState(StateID, CurrentCellContents)
    
    If Not (clsTemp Is Nothing) Then
        ' The state has been found.
        DFAStateExists = True
    Else
        ' The state hasn't been found.
        DFAStateExists = False
    End If
End Function


'
' Function GetDFAStateIDs
'
' Purpose:
'   Returns a collection of unique DFA state IDs which are found in the DFA state collection.
'
' Returns:
'   A collection of unique DFA state IDs which are found in the DFA state collection.
'
Public Function GetDFAStateIDs() As Collection
    Dim colRetStates As Collection
    Dim clsCurrentState As clsDFAState
    Dim clsTemp As clsDFAState
    
    Set colRetStates = New Collection
    
    On Error Resume Next
    
    For Each clsCurrentState In mcolDFAStates
        ' Add the current state ID only if it wasn't already added to the collection.
        Set clsTemp = colRetStates(clsCurrentState.StateID)
        If (clsTemp Is Nothing) Then _
            colRetStates.Add clsCurrentState.StateID, clsCurrentState.StateID
    Next clsCurrentState
    
    Set GetDFAStateIDs = colRetStates
End Function


'
' Function GetDFAStates
'
' Purpose:
'   Returns all of the DFA states which have the same StateID property.
'
' Parameters:
'   StateID : The ID of the states.
'
' Returns:
'   A collection of the DFA states which have the same StateID property.
'
Public Function GetDFAStates(StateID As String) As Collection
    Dim colStates As Collection
    Dim clsCurrentState As clsDFAState
    
    Set colStates = New Collection
    
    For Each clsCurrentState In mcolDFAStates
        If (clsCurrentState.StateID = StateID) Then
            colStates.Add clsCurrentState, clsCurrentState.CurrentCellContents
        End If
    Next clsCurrentState
    
    Set GetDFAStates = colStates
End Function


'
' Function GetDFAState
'
' Purpose:
'   Returns a DFA state class according to its StateID and its CurrentCellValue properties.
'
' Parameters:
'   StateID : The ID of the state.
'   CurrentCellContents : The current cell value of the state.
'
' Returns:
'   The DFA state which has the indicated StateID and CurrentCellValue values.
'
Public Function GetDFAState(StateID As String, CurrentCellContents As String) As clsDFAState
    Dim clsCurrentState As clsDFAState
    
    For Each clsCurrentState In mcolDFAStates
        ' Found a matching state.
        If ((clsCurrentState.StateID = StateID) And _
            (clsCurrentState.CurrentCellContents = CurrentCellContents)) Then _
            Set GetDFAState = clsCurrentState: Exit Function
    Next clsCurrentState
    
    ' State no found.
    Set GetDFAState = Nothing
    
End Function


'
' Public Subs
'


'
' Sub AddTapeCell
'
' Purpose:
'   Adds a new tape cell at the end of the tape.
'
' Parameters:
'   NewCell : The value of the new cell.
'
Public Sub AddTapeCell(NewCell As String)
    ' Invalid cell value (since textual cell values option is not set).
    If ((Not mblnTextualTapeCellValues) And (NewCell <> "0") And (NewCell <> "1")) Then _
        RaiseEvent ErrorRaised(deInvalidCellValue, "Invalid cell value entered: " & NewCell, "clsDFA.AddTapeCell"): Exit Sub
    
    ReDim Preserve mstrTapeCells(LBound(mstrTapeCells) To UBound(mstrTapeCells) + 1)
    mstrTapeCells(UBound(mstrTapeCells)) = NewCell
    
    ReDim Preserve mblnTapeCellsBreakpoints(LBound(mblnTapeCellsBreakpoints) To UBound(mblnTapeCellsBreakpoints) + 1)
End Sub


'
' Sub DeleteTapeCell
'
' Purpose:
'   Deletes a specific tape cell.
'
' Parameters:
'   Index : The index of the cell to be deleted.
'
Public Sub DeleteTapeCell(Index As Integer)
    Dim i As Integer
    
    ' Invalid index.
    If ((Index < LBound(mstrTapeCells)) Or (Index > UBound(mstrTapeCells))) Then _
        RaiseEvent ErrorRaised(deInvalidIndex, "Invalid index entered: " & Index, "clsDFA.DeleteTapeCell"): Exit Sub
    
    If (Index = UBound(mstrTapeCells)) Then
        ReDim Preserve mstrTapeCells(LBound(mstrTapeCells) To UBound(mstrTapeCells) - 1)
        ReDim Preserve mblnTapeCellsBreakpoints(LBound(mblnTapeCellsBreakpoints) To UBound(mblnTapeCellsBreakpoints) - 1)
        
        ' Change the current tape position if it's bigger than the new tape size.
        If (mintCurrentTapePosition > Index - 1) Then _
            mintCurrentTapePosition = Index - 1
    Else
        For i = Index To UBound(mstrTapeCells) - 1
            mstrTapeCells(i) = mstrTapeCells(i + 1)
            mblnTapeCellsBreakpoints(i) = mblnTapeCellsBreakpoints(i + 1)
        Next i
        
        ReDim Preserve mstrTapeCells(LBound(mstrTapeCells) To UBound(mstrTapeCells) - 1)
        ReDim Preserve mblnTapeCellsBreakpoints(LBound(mblnTapeCellsBreakpoints) To UBound(mblnTapeCellsBreakpoints) - 1)
    End If
End Sub


'
' Sub AddState
'
' Purpose:
'   Adds a new state to the state collection.
'
' Parameters:
'   NewState : The state to be added.
'
Public Sub AddState(NewState As clsDFAState)
    Dim clsTemp As clsDFAState
    
    On Error Resume Next
    
    ' The simulation mustn't be running in order to call this sub.
    If (mssSimulationStatus = ssRunning) Then _
        RaiseEvent ErrorRaised(deSubCannotRunWhileSimulatorRunning, "The sub cannot run while the DFA simulator is running", "clsDFA.AddState"): Exit Sub
    
    ' Invalid state.
    If ((NewState.StateID = "") Or (NewState.NewStateID = "") Or _
        (NewState.CurrentCellContents = "") Or (NewState.NewCellContents = "")) Then _
        RaiseEvent ErrorRaised(deInvalidState, "Invalid DFA State entered", "clsDFA.AddState"): Exit Sub
    
    ' Create a new collection if necessary.
    If (mcolDFAStates Is Nothing) Then _
        Set mcolDFAStates = New Collection

    ' See if the new state already exists.
    If (DFAStateExists(NewState.StateID, NewState.CurrentCellContents) = True) Then _
        RaiseEvent ErrorRaised(deStateAlreadyExists, "The DFA state being added already exists", "clsDFA.AddState"): Exit Sub
    
    ' Add the new state to the state collection.
    mcolDFAStates.Add NewState
End Sub


'
' Sub EditState
'
' Purpose:
'   Changes the ID of a state
'
' Parameters:
'   OldStateID : The ID of the old, previous state.
'   NewStateID : The new state ID.
'
' Note:
'   This sub changes all references of the old state ID to the new state ID.
'
Public Sub EditState(OldStateID As String, NewStateID As String)
    Dim clsCurrentState As clsDFAState
    
    For Each clsCurrentState In mcolDFAStates
        If (clsCurrentState.StateID = OldStateID) Then _
            clsCurrentState.StateID = NewStateID
        If (clsCurrentState.NewStateID = OldStateID) Then _
            clsCurrentState.NewStateID = NewStateID
    Next clsCurrentState

    If (mstrInitialStateID = OldStateID) Then _
        mstrInitialStateID = NewStateID
    If (mstrCurrentStateID = OldStateID) Then _
        mstrCurrentStateID = NewStateID
End Sub



'
' Sub DeleteState
'
' Purpose:
'   Deletes an existing state from the state collection.
'
' Parameters:
'   StateID : The ID of the state to be deleted.
'   CurrentCellContents : The current cell value of the state to be deleted.
'
Public Sub DeleteState(StateID As String, CurrentCellContents As String)
    Dim clsCurrentState As clsDFAState
    Dim i As Integer
    
    On Error Resume Next
    
    ' The simulation mustn't be running in order to call this sub.
    If (mssSimulationStatus = ssRunning) Then _
        RaiseEvent ErrorRaised(deSubCannotRunWhileSimulatorRunning, "The sub cannot run while the DFA simulator is running", "clsDFA.DeleteState"): Exit Sub
    
    ' Invalid state.
    If ((StateID = "") Or (CurrentCellContents = "")) Then _
        RaiseEvent ErrorRaised(deInvalidState, "Invalid DFA State entered", "clsDFA.DeleteState"): Exit Sub
    
    ' Find the state matching the current cell value.
    For i = 1 To mcolDFAStates.Count
        Set clsCurrentState = mcolDFAStates(i)
        If ((clsCurrentState.StateID = StateID) And (clsCurrentState.CurrentCellContents = CurrentCellContents)) Then _
            mcolDFAStates.Remove i: Exit Sub
    Next i
    
    ' If we've reached this far, this must mean that the state hasn't been found.
    RaiseEvent ErrorRaised(deDFAStateNotFound, "The DFA state wasn't found", "clsDFA.DeleteState"): Exit Sub
End Sub


'
' Sub DeleteStates
'
' Purpose:
'   Deletes states which have the specified StateID from the state collection.
'
' Parameters:
'   StateID : The State ID of the states to be deleted.
'
Public Sub DeleteStates(StateID As String)
    Dim clsCurrentState As clsDFAState
    Dim i As Integer
    
    For i = 1 To mcolDFAStates.Count
        Set clsCurrentState = mcolDFAStates(i)
        If (clsCurrentState.StateID = StateID) Then
            ' Found a state with a matching state ID - delete the current state.
            mcolDFAStates.Remove i
        End If
    Next i
End Sub


'
' Sub NewDFA
'
' Purpose:
'   Creates a new DFA within the current class (deletes the old DFA).
'
Public Sub NewDFA()
    Dim clsCurrentDFAState As clsDFAState

    ' Reset the DFA simulation first.
    ResetSimulation
    
    ' Remove all DFA states.
    Do While (mcolDFAStates.Count > 0)
        Set clsCurrentDFAState = mcolDFAStates(1)
        Set clsCurrentDFAState = Nothing
        
        mcolDFAStates.Remove 1
    Loop
    
    mstrInitialStateID = ""
    mstrCurrentStateID = ""
    
    ' Reset the DFA properties.
    mstrDFAName = ""
    mstrDFAAuthor = ""
    mstrDFADescription = ""
    
    mblnTextualTapeCellValues = False
    mblnReadHeadCanMoveSeveralSteps = False
    mblnReadHeadCanStayAtPosition = False
    mblnAllowAcceptingStates = False
End Sub


'
' Sub LoadDFA
'
' Purpose:
'   Loads a DFA from a file.
'
' Parameters:
'   FileName : The file in which the DFA is located.
'
' Note:
'   This subs overwrites the previous DFA states when loading
'   a new DFA from file.
'
Public Sub LoadDFA(Filename As String)
    Dim intDFAFile As Integer
    Dim clsCurrentState As clsDFAState
    Dim intDFAStateCount As Integer
    
    Dim bytTemp As Byte, blnTemp As Boolean, strTemp As String, intTemp As Integer, i As Integer

    ' The simulation mustn't be running in order to call this sub.
    If (mssSimulationStatus = ssRunning) Then _
        RaiseEvent ErrorRaised(deSubCannotRunWhileSimulatorRunning, "The sub cannot run while the DFA simulator is running", "clsDFA.LoadDFA"): Exit Sub
    
    ' File doesn't exist.
    If (Dir$(Filename) = "") Then _
        RaiseEvent ErrorRaised(deFileDoesntExist, "The file '" & Filename & "' doesn't exist", "clsDFA.LoadDFA"): Exit Sub
    ' Not a valid DFA file.
    If (FileLen(Filename) < 9) Then _
        RaiseEvent ErrorRaised(deInvalidDFAFile, "The file '" & Filename & "' isn't a valid DFA file", "clsDFA.LoadDFA"): Exit Sub
    
    ' Open the DFA file.
    
    intDFAFile = FreeFile
    
    Open Filename For Binary As intDFAFile
    
    ' Read the DFA header.
    
    
    ' The identification string.
    
    Get intDFAFile, , bytTemp
    ' Not a valid DFA file.
    If (bytTemp <> Asc("D")) Then _
        RaiseEvent ErrorRaised(deInvalidDFAFile, "The file '" & Filename & "' isn't a valid DFA file", "clsDFA.LoadDFA"): Exit Sub
    Get intDFAFile, , bytTemp
    ' Not a valid DFA file.
    If (bytTemp <> Asc("F")) Then _
        RaiseEvent ErrorRaised(deInvalidDFAFile, "The file '" & Filename & "' isn't a valid DFA file", "clsDFA.LoadDFA"): Exit Sub
    Get intDFAFile, , bytTemp
    ' Not a valid DFA file.
    If (bytTemp <> Asc("A")) Then _
        RaiseEvent ErrorRaised(deInvalidDFAFile, "The file '" & Filename & "' isn't a valid DFA file", "clsDFA.LoadDFA"): Exit Sub
        
    ' Read the DFA's name, description and author.
    mstrDFAName = ReadString(intDFAFile)
    mstrDFADescription = ReadString(intDFAFile)
    mstrDFAAuthor = ReadString(intDFAFile)

    ' Read the advanced options.
    Get intDFAFile, , mblnTextualTapeCellValues
    Get intDFAFile, , mblnReadHeadCanMoveSeveralSteps
    Get intDFAFile, , mblnReadHeadCanStayAtPosition
    Get intDFAFile, , mblnAllowAcceptingStates
    
    ' Read the ID of the initial state.
    mstrInitialStateID = ReadString(intDFAFile)
    
    ' Read the total no. of states.
    Get intDFAFile, , intDFAStateCount
    
    ' Not a valid DFA file.
    If ((intDFAStateCount <= 0) Or (mstrInitialStateID = "") Or (EOF(intDFAFile))) Then _
        RaiseEvent ErrorRaised(deInvalidDFAFile, "The file '" & Filename & "' isn't a valid DFA file", "clsDFA.LoadDFA"): Exit Sub
    
    ' Remove previous DFA states.
    Do While (mcolDFAStates.Count > 0)
        Set clsCurrentState = mcolDFAStates(1)
        Set clsCurrentState = Nothing
        mcolDFAStates.Remove 1
    Loop
    
    ' Read all of the DFA states.
    
    For i = 1 To intDFAStateCount
    
        ' Not a valid DFA file.
        If (EOF(intDFAFile)) Then _
            RaiseEvent ErrorRaised(deInvalidDFAFile, "The file '" & Filename & "' isn't a valid DFA file", "clsDFA.LoadDFA"): Exit Sub
    
        Set clsCurrentState = New clsDFAState
        
        ' Read the current state ID.
        clsCurrentState.StateID = ReadString(intDFAFile)
        ' Read the current cell contents.
        If (mblnTextualTapeCellValues) Then
            clsCurrentState.CurrentCellContents = ReadString(intDFAFile)
        Else
            Get intDFAFile, , bytTemp: clsCurrentState.CurrentCellContents = CStr(bytTemp)
        End If
        ' Read the new cell contents.
        If (mblnTextualTapeCellValues) Then
            clsCurrentState.NewCellContents = ReadString(intDFAFile)
        Else
            Get intDFAFile, , bytTemp: clsCurrentState.NewCellContents = CStr(bytTemp)
        End If
        ' Read the new direction.
        Get intDFAFile, , intTemp: clsCurrentState.NewDirection = intTemp
        ' Read the new state ID.
        clsCurrentState.NewStateID = ReadString(intDFAFile)
        ' See if this state is an accepting state (if the "Allow accepting states" option is set).
        If (mblnAllowAcceptingStates) Then _
            Get intDFAFile, , blnTemp: clsCurrentState.AcceptingState = blnTemp
        ' Read the state's comment.
        clsCurrentState.Comment = ReadString(intDFAFile)
        ' Read the state's breakpoint status.
        blnTemp = clsCurrentState.BreakPoint
        Get intDFAFile, , blnTemp
    
        ' Invalid state.
        If ((clsCurrentState.StateID = "") Or (clsCurrentState.NewStateID = "") Or _
            (clsCurrentState.CurrentCellContents = "") Or (clsCurrentState.NewCellContents = "")) Then _
            RaiseEvent ErrorRaised(deInvalidState, "Invalid DFA State read", "clsDFA.LoadDFA"): Exit Sub
        
        ' Add the new state to the state collection.
        mcolDFAStates.Add clsCurrentState
    Next i
    
    Close intDFAFile

    ' Reset the DFA simulation.
    ResetSimulation
End Sub


'
' Sub SaveDFA
'
' Purpose:
'   Saves the current DFA onto a file.
'
' Parameters:
'   FileName : The file in which the DFA will be saved.
'
Public Sub SaveDFA(Filename As String)
    Dim intDFAFile As Integer
    Dim clsCurrentState As clsDFAState
    
    ' Overwrite the file if it already exists.
    If (Dir$(Filename) <> "") Then _
        Kill Filename
    
    ' Create the DFA file.
    
    intDFAFile = FreeFile
    
    Open Filename For Binary As intDFAFile
    
    ' Write the DFA header.
    
    ' Write some sort of identification string.
    Put intDFAFile, , CByte(Asc("D"))
    Put intDFAFile, , CByte(Asc("F"))
    Put intDFAFile, , CByte(Asc("A"))
    
    ' Write the DFA's name, description and author.
    WriteString intDFAFile, mstrDFAName
    WriteString intDFAFile, mstrDFADescription
    WriteString intDFAFile, mstrDFAAuthor
    
    ' Write the advanced options.
    Put intDFAFile, , mblnTextualTapeCellValues
    Put intDFAFile, , mblnReadHeadCanMoveSeveralSteps
    Put intDFAFile, , mblnReadHeadCanStayAtPosition
    Put intDFAFile, , mblnAllowAcceptingStates
    
    ' Write the ID of the initial state.
    WriteString intDFAFile, mstrInitialStateID
    
    ' Write the total no. of states.
    Put intDFAFile, , CInt(mcolDFAStates.Count)
    
    ' Write all of the DFA states.
    
    For Each clsCurrentState In mcolDFAStates
        ' Write the current state ID.
        WriteString intDFAFile, clsCurrentState.StateID
        ' Write the current cell contents.
        If (mblnTextualTapeCellValues) Then
            WriteString intDFAFile, clsCurrentState.CurrentCellContents
        Else
            Put intDFAFile, , CByte(clsCurrentState.CurrentCellContents)
        End If
        ' Write the new cell contents.
        If (mblnTextualTapeCellValues) Then
            WriteString intDFAFile, clsCurrentState.NewCellContents
        Else
            Put intDFAFile, , CByte(clsCurrentState.NewCellContents)
        End If
        ' Write the new direction.
        Put intDFAFile, , CInt(clsCurrentState.NewDirection)
        ' Write the new state ID.
        WriteString intDFAFile, clsCurrentState.NewStateID
        ' Write if this state is an accepting state (if the "Allow accepting states" option is set).
        If (mblnAllowAcceptingStates) Then _
            Put intDFAFile, , clsCurrentState.AcceptingState
        ' Write the state's comment.
        WriteString intDFAFile, clsCurrentState.Comment
        ' Write the state's breakpoint status.
        Put intDFAFile, , clsCurrentState.BreakPoint
    Next clsCurrentState
    
    Close intDFAFile
End Sub


'
' Sub LoadTextualDFA
'
' Purpose:
'   Loads a Textual DFA from a file (the DFA file format is different - it's textual,
'   not binary, so it can be edited by hand using a simple text editor).
'
' Parameters:
'   FileName : The file in which the DFA is located.
'
' Note:
'   This subs overwrites the previous DFA states when loading
'   a new DFA from file.
'
Public Sub LoadTextualDFA(Filename As String)
    Dim intDFAFile As Integer
    Dim strDFAFile As String
    
    ' The simulation mustn't be running in order to call this sub.
    If (mssSimulationStatus = ssRunning) Then _
        RaiseEvent ErrorRaised(deSubCannotRunWhileSimulatorRunning, "The sub cannot run while the DFA simulator is running", "clsDFA.LoadTextualDFA"): Exit Sub
    
    ' File doesn't exist.
    If (Dir$(Filename) = "") Then _
        RaiseEvent ErrorRaised(deFileDoesntExist, "The file '" & Filename & "' doesn't exist", "clsDFA.LoadTextualDFA"): Exit Sub
    ' Not a valid DFA file.
    If (FileLen(Filename) = 0) Then _
        RaiseEvent ErrorRaised(deInvalidDFAFile, "The file '" & Filename & "' isn't a valid DFA file", "clsDFA.LoadTextualDFA"): Exit Sub
    
    ' Open the textual DFA file.
    
    intDFAFile = FreeFile
    
    Open Filename For Input As intDFAFile
        ' Read the contents of the entire file.
        strDFAFile = Input(LOF(intDFAFile), intDFAFile)
    Close intDFAFile
    
    ' Read the contents of the textual DFA file.
    LoadDFAFromText (strDFAFile)
End Sub


'
' Sub LoadDFAFromText
'
' Purpose:
'   Loads a DFA from a given text segment.
'
' Parameters:
'   Text : The text in which the DFA is defined.
'
' Note:
'   This sub overwrites any existing DFA states while loading the new DFA.
'
Public Sub LoadDFAFromText(ByVal Text As String)
    Dim clsCurrentState As clsDFAState

    Dim strCurrentLine As String, strOriginalLine As String
    Dim strPreviousStateID As String

    Dim intTemp As Integer, i As Integer, c As Integer
    

    ' The simulation mustn't be running in order to call this sub.
    If (mssSimulationStatus = ssRunning) Then _
        RaiseEvent ErrorRaised(deSubCannotRunWhileSimulatorRunning, "The sub cannot run while the DFA simulator is running", "clsDFA.LoadDFAFromText"): Exit Sub
    
    ' Not a valid DFA text.
    If (Trim$(Text) = "") Then _
        RaiseEvent ErrorRaised(deInvalidDFAFile, "The text isn't a valid DFA (it's empty).", "clsDFA.LoadDFAFromText"): Exit Sub
    
    ' Reset the DFA properties.
    
    mstrDFAName = ""
    mstrDFADescription = ""
    mstrDFAAuthor = ""
    
    mblnTextualTapeCellValues = False
    mblnReadHeadCanMoveSeveralSteps = False
    mblnReadHeadCanStayAtPosition = False
    mblnAllowAcceptingStates = False
    
    mstrInitialStateID = ""
    
    ' Remove previous DFA states.
    Do While (mcolDFAStates.Count > 0)
        Set clsCurrentState = mcolDFAStates(1)
        Set clsCurrentState = Nothing
        mcolDFAStates.Remove 1
    Loop
    
    ' Read the DFA header.
    
    Do While (Text <> "")
        strCurrentLine = ""
        
        ' Skip any blank\comment lines.
        Do While (((strCurrentLine = "") Or (Left$(strCurrentLine, 1) = "'")) And (Text <> ""))
            i = InStr(1, Text, vbCrLf)
            If (i = 0) Then i = Len(Text) + 1
            
            strCurrentLine = Trim$(Left$(Text, i - 1))
            
            If (i > Len(Text)) Then
                Text = ""
            Else
                Text = Mid$(Text, i + 2)
            End If
        Loop
        
        strOriginalLine = strCurrentLine
        
        If (LCase$(Left$(strCurrentLine, 8)) = "dfa name") Then
            ' The name of the DFA.
            
            strCurrentLine = Trim$(Mid$(strCurrentLine, 9))
            
            If (Left$(strCurrentLine, 1) <> "=") Then _
                RaiseEvent ErrorRaised(deInvalidDFAProperty, "The current DFA property, 'DFA Name', has no value set.", "clsDFA.LoadDFAFromText"): Exit Sub
            
            mstrDFAName = Trim$(Mid$(strCurrentLine, 2))
            
        ElseIf (LCase$(Left$(strCurrentLine, 15)) = "dfa description") Then
            ' The DFA description.
            
            strCurrentLine = Trim$(Mid$(strCurrentLine, 16))
            
            If (Left$(strCurrentLine, 1) <> "=") Then _
                RaiseEvent ErrorRaised(deInvalidDFAProperty, "The current DFA property, 'DFA Description', has no value set.", "clsDFA.LoadDFAFromText"): Exit Sub
            
            mstrDFADescription = Trim$(Mid$(strCurrentLine, 2))
            
            mstrDFADescription = ReplaceString(mstrDFADescription, "\\", "\")
            mstrDFADescription = ReplaceString(mstrDFADescription, "\n", vbCrLf)
            mstrDFADescription = ReplaceString(mstrDFADescription, "\t", vbTab)
            
        ElseIf (LCase$(Left$(strCurrentLine, 10)) = "dfa author") Then
            ' The author of the DFA.
            
            strCurrentLine = Trim$(Mid$(strCurrentLine, 11))
            
            If (Left$(strCurrentLine, 1) <> "=") Then _
                RaiseEvent ErrorRaised(deInvalidDFAProperty, "The current DFA property, 'DFA Author', has no value set.", "clsDFA.LoadDFAFromText"): Exit Sub
            
            mstrDFAAuthor = Trim$(Mid$(strCurrentLine, 2))
            
        ElseIf (LCase$(Left$(strCurrentLine, 18)) = "textual tape cells") Then
            ' Textual Tape Cell Values option.
            
            strCurrentLine = Trim$(Mid$(strCurrentLine, 19))
            
            If (Left$(strCurrentLine, 1) <> "=") Then _
                RaiseEvent ErrorRaised(deInvalidDFAProperty, "The current DFA property, 'Textual Tape Cells', has no value set (either True\T or False\F).", "clsDFA.LoadDFAFromText"): Exit Sub
            
            If ((LCase$(Trim$(Mid$(strCurrentLine, 2))) = "true") Or _
                (LCase$(Trim$(Mid$(strCurrentLine, 2))) = "t")) Then
                mblnTextualTapeCellValues = True
            ElseIf ((LCase$(Trim$(Mid$(strCurrentLine, 2))) = "false") Or _
                (LCase$(Trim$(Mid$(strCurrentLine, 2))) = "f")) Then
                mblnTextualTapeCellValues = False
            Else
                RaiseEvent ErrorRaised(deInvalidDFAProperty, "The current DFA property, 'Textual Tape Cells', has an invalid value set - '" & Mid$(strCurrentLine, 2) & "'." & vbCrLf & "It can only be either True\T or False\F.", "clsDFA.LoadDFAFromText")
                Exit Sub
            End If
            
        ElseIf (LCase$(Left$(strCurrentLine, 32)) = "read head can move several steps") Then
            ' Read Head Can Move Several Steps option.
            
            strCurrentLine = Trim$(Mid$(strCurrentLine, 33))
            
            If (Left$(strCurrentLine, 1) <> "=") Then _
                RaiseEvent ErrorRaised(deInvalidDFAProperty, "The current DFA property, 'Read Head Can Move Several Steps', has no value set (either True\T or False\F).", "clsDFA.LoadDFAFromText"): Exit Sub
            
            If ((LCase$(Trim$(Mid$(strCurrentLine, 2))) = "true") Or _
                (LCase$(Trim$(Mid$(strCurrentLine, 2))) = "t")) Then
                mblnReadHeadCanMoveSeveralSteps = True
            ElseIf ((LCase$(Trim$(Mid$(strCurrentLine, 2))) = "false") Or _
                (LCase$(Trim$(Mid$(strCurrentLine, 2))) = "f")) Then
                mblnReadHeadCanMoveSeveralSteps = False
            Else
                RaiseEvent ErrorRaised(deInvalidDFAProperty, "The current DFA property, 'Read Head Can Move Several Steps', has an invalid value set - '" & Mid$(strCurrentLine, 2) & "'." & vbCrLf & "It can only be either True\T or False\F.", "clsDFA.LoadDFAFromText")
                Exit Sub
            End If
            
        ElseIf (LCase$(Left$(strCurrentLine, 35)) = "read head can stay at same position") Then
            ' Read Head Can Stay At The Same Position option.
            
            strCurrentLine = Trim$(Mid$(strCurrentLine, 36))
            
            If (Left$(strCurrentLine, 1) <> "=") Then _
                RaiseEvent ErrorRaised(deInvalidDFAProperty, "The current DFA property, 'Read Head Can Stay At Same Position', has no value set (either True\T or False\F).", "clsDFA.LoadDFAFromText"): Exit Sub
            
            If ((LCase$(Trim$(Mid$(strCurrentLine, 2))) = "true") Or _
                (LCase$(Trim$(Mid$(strCurrentLine, 2))) = "t")) Then
                mblnReadHeadCanStayAtPosition = True
            ElseIf ((LCase$(Trim$(Mid$(strCurrentLine, 2))) = "false") Or _
                (LCase$(Trim$(Mid$(strCurrentLine, 2))) = "f")) Then
                mblnReadHeadCanStayAtPosition = False
            Else
                RaiseEvent ErrorRaised(deInvalidDFAProperty, "The current DFA property, 'Read Head Can Stay At Same Position', has an invalid value set - '" & Mid$(strCurrentLine, 2) & "'." & vbCrLf & "It can only be either True\T or False\F.", "clsDFA.LoadDFAFromText")
                Exit Sub
            End If
            
        ElseIf (LCase$(Left$(strCurrentLine, 21)) = "allow accepting rules") Then
            ' Allow Accepting States option.
            
            strCurrentLine = Trim$(Mid$(strCurrentLine, 22))
            
            If (Left$(strCurrentLine, 1) <> "=") Then _
                RaiseEvent ErrorRaised(deInvalidDFAProperty, "The current DFA property, 'Allow Accepting Rules', has no value set (either True\T or False\F).", "clsDFA.LoadDFAFromText"): Exit Sub
            
            If ((LCase$(Trim$(Mid$(strCurrentLine, 2))) = "true") Or _
                (LCase$(Trim$(Mid$(strCurrentLine, 2))) = "t")) Then
                mblnAllowAcceptingStates = True
            ElseIf ((LCase$(Trim$(Mid$(strCurrentLine, 2))) = "false") Or _
                (LCase$(Trim$(Mid$(strCurrentLine, 2))) = "f")) Then
                mblnAllowAcceptingStates = False
            Else
                RaiseEvent ErrorRaised(deInvalidDFAProperty, "The current DFA property, 'Allow Accepting Rules', has an invalid value set - '" & Mid$(strCurrentLine, 2) & "'." & vbCrLf & "It can only be either True\T or False\F.", "clsDFA.LoadDFAFromText")
                Exit Sub
            End If
            
        ElseIf (LCase$(Left$(strCurrentLine, 13)) = "initial state") Then
            ' Initial State ID.
            
            strCurrentLine = Trim$(Mid$(strCurrentLine, 14))
            
            If (Left$(strCurrentLine, 1) <> "=") Then _
                RaiseEvent ErrorRaised(deInvalidDFAProperty, "The current DFA property, 'Initial State', has no value set (it must be the name of an existing DFA state).", "clsDFA.LoadDFAFromText"): Exit Sub
            
            mstrInitialStateID = Trim$(Mid$(strCurrentLine, 2))
            
            If (mstrInitialStateID = "") Then _
                RaiseEvent ErrorRaised(deNoInitialDFAState, "The initial state ID is missing!", "clsDFA.LoadDFAFromText"): Exit Sub
            
        ElseIf (LCase$(Left$(strCurrentLine, 8)) = "dfa code") Then
            ' Marks the beginning of the DFA program code (and the end of the DFA header).
            
            If ((Trim$(Mid$(strCurrentLine, 9)) <> ":") And (Trim$(Mid$(strCurrentLine, 9)) <> "")) Then _
                RaiseEvent ErrorRaised(deInvalidDFAStatement, "The current statement, '" & strCurrentLine & "' is invalid. Did you mean to write 'DFA Code:'?", "clsDFA.LoadDFAFromText"): Exit Sub
            
            strCurrentLine = Trim$(Mid$(strCurrentLine, 9))
            
            Exit Do
        ElseIf (Text <> "") Then
            ' Invalid DFA property.
            
            RaiseEvent ErrorRaised(deInvalidDFAProperty, "The current DFA property, '" & strCurrentLine & "', is an invalid DFA property.", "clsDFA.LoadDFAFromText")
            Exit Sub
        End If
    Loop
    
    If (mstrInitialStateID = "") Then _
        RaiseEvent ErrorRaised(deNoInitialDFAState, "The initial state ID is missing!", "clsDFA.LoadDFAFromText"): Exit Sub
    If (Text = "") Then _
        RaiseEvent ErrorRaised(deNoDFAStates, "The DFA rules are missing!", "clsDFA.LoadDFAFromText"): Exit Sub


    Set clsCurrentState = New clsDFAState
    strPreviousStateID = mstrInitialStateID

    ' Read all of the DFA states.
    
    Do While (Text <> "")
        strCurrentLine = ""
        
        ' Skip any blank lines.
        Do While ((strCurrentLine = "") And (Text <> ""))
            i = InStr(1, Text, vbCrLf)
            If (i = 0) Then i = Len(Text) + 1
            
            strCurrentLine = Trim$(Left$(Text, i - 1))

            If (i > Len(Text)) Then
                Text = ""
            Else
                Text = Mid$(Text, i + 2)
            End If
        Loop
        
        strOriginalLine = strCurrentLine

        If (Left$(strCurrentLine, 1) = "'") Then
            ' A Comment.
            clsCurrentState.Comment = clsCurrentState.Comment & IIf(clsCurrentState.Comment <> "", " ", "") & Trim$(Mid$(strCurrentLine, 2))
            
        ElseIf (strCurrentLine <> "") Then
            ' A Rule definition.
            
            ' Illegal Rule definition.
            If ((Left$(strCurrentLine, 1) <> """") And (Left$(strCurrentLine, 1) <> ",") And _
                (LCase$(Left$(strCurrentLine, 6)) <> "break:")) Then _
                RaiseEvent ErrorRaised(deInvalidDFAStatement, "The current statement, '" & strCurrentLine & "', isn't a valid rule definition.", "clsDFA.LoadDFAFromText"): Exit Sub
            
            
            ' Parse the rule definition.
            
            If (LCase$(Left$(strCurrentLine, 6)) = "break:") Then
                ' The current rule if a breakpoint.
                clsCurrentState.BreakPoint = True
                
                strCurrentLine = Trim$(Mid$(strCurrentLine, 7))
            End If
            
            ' Current State ID (Optional).
            If (Left$(strCurrentLine, 1) = """") Then
                i = InStr(2, strCurrentLine, """")
                
                ' Illegal Rule definition.
                If (i = 0) Then _
                    RaiseEvent ErrorRaised(deInvalidState, "The current rule's ('" & strOriginalLine & "') Current State ID property is missing its closing quotation mark.", "clsDFA.LoadDFAFromText"): Exit Sub
                
                clsCurrentState.StateID = Trim$(Mid$(strCurrentLine, 2, i - 2))
                
                i = InStr(i, strCurrentLine, ",")
                
                ' Illegal Rule definition.
                If (i = 0) Then _
                    RaiseEvent ErrorRaised(deInvalidState, "The current rule, '" & strOriginalLine & "', is missing the Current and New Cell Contents properties.", "clsDFA.LoadDFAFromText"): Exit Sub
                
                strCurrentLine = Trim$(Mid$(strCurrentLine, i + 1))
                
            ElseIf (Left$(strCurrentLine, 1) = ",") Then
                ' Empty State ID ( = Previous State ID).
                clsCurrentState.StateID = strPreviousStateID
                
                strCurrentLine = Trim$(Mid$(strCurrentLine, 2))
            End If
        
            ' Current Cell Contents.
            If (((Left$(strCurrentLine, 1) = """") And (mblnTextualTapeCellValues = True)) Or _
                ((Left$(strCurrentLine, 1) <> ",") And (strCurrentLine <> "") And (mblnTextualTapeCellValues = False))) Then
                
                If (mblnTextualTapeCellValues = True) Then
                    ' Textual Tape Cell Values must be surrounded by quotation marks.
                    i = InStr(2, strCurrentLine, """")
                
                    ' Illegal Rule definition.
                    If (i = 0) Then _
                        RaiseEvent ErrorRaised(deInvalidState, "The current rule's ('" & strOriginalLine & "') Current Cell Contents property is missing its closing quotation mark.", "clsDFA.LoadDFAFromText"): Exit Sub
                Else
                    i = InStr(1, strCurrentLine, ",")
                
                    ' Illegal Rule definition.
                    If (i = 0) Then _
                        RaiseEvent ErrorRaised(deInvalidState, "The current rule, '" & strOriginalLine & "', is missing the New Cell Contents property.", "clsDFA.LoadDFAFromText"): Exit Sub
                End If
                
                If (mblnTextualTapeCellValues = True) Then
                    clsCurrentState.CurrentCellContents = Mid$(strCurrentLine, 2, i - 2)
                Else
                    clsCurrentState.CurrentCellContents = Trim$(Left$(strCurrentLine, i - 1))
                
                    If ((clsCurrentState.CurrentCellContents <> "0") And _
                        (clsCurrentState.CurrentCellContents <> "1")) Then _
                        RaiseEvent ErrorRaised(deInvalidDFAStateValue, "The current rule's ('" & strOriginalLine & "') Current Cell Contents property has an invalid value: '" & clsCurrentState.NewCellContents & "'." & vbCrLf & "It can only be either 0 or 1.", "clsDFA.LoadDFAFromText"): Exit Sub
                End If
                
                strCurrentLine = Trim$(Mid$(strCurrentLine, i + 1))
                
                If (mblnTextualTapeCellValues = True) Then
                    ' Search for the comma.
                    i = InStr(1, strCurrentLine, ",")
                    
                    If (i = 0) Then _
                        RaiseEvent ErrorRaised(deInvalidState, "The current rule, '" & strOriginalLine & "', is missing the New Cell Contents property.", "clsDFA.LoadDFAFromText"): Exit Sub
                    
                    strCurrentLine = Trim$(Mid$(strCurrentLine, i + 1))
                End If
                
            Else
                ' This rule definition parameter isn't optional.
                RaiseEvent ErrorRaised(deInvalidState, "The current rule, '" & strOriginalLine & "', is missing the Current and New Cell Contents properties.", "clsDFA.LoadDFAFromText")
                Exit Sub
            End If
            
        
            ' New Cell Contents.
            If (((Left$(strCurrentLine, 1) = """") And (mblnTextualTapeCellValues = True)) Or _
                ((Left$(strCurrentLine, 1) <> ",") And (strCurrentLine <> "") And (mblnTextualTapeCellValues = False))) Then
                
                If (mblnTextualTapeCellValues = True) Then
                    ' Textual Tape Cell Values must be surrounded by quotation marks.
                    i = InStr(2, strCurrentLine, """")
                    
                    ' Illegal Rule definition.
                    If (i = 0) Then _
                        RaiseEvent ErrorRaised(deInvalidState, "The current rule's ('" & strOriginalLine & "') New Cell Contents property is missing its closing quotation mark.", "clsDFA.LoadDFAFromText"): Exit Sub
                Else
                    i = InStr(1, strCurrentLine, ",")
                    
                    If (i = 0) Then _
                        i = Len(strCurrentLine) + 1
                End If
                
                If (mblnTextualTapeCellValues = True) Then
                    clsCurrentState.NewCellContents = Mid$(strCurrentLine, 2, i - 2)
                Else
                    clsCurrentState.NewCellContents = Trim$(Left$(strCurrentLine, i - 1))
                    
                    If ((clsCurrentState.NewCellContents <> "0") And _
                        (clsCurrentState.NewCellContents <> "1")) Then _
                        RaiseEvent ErrorRaised(deInvalidDFAStateValue, "The current rule's ('" & strOriginalLine & "') New Cell Contents property has an invalid value: '" & clsCurrentState.NewCellContents & "'." & vbCrLf & "It can only be either 0 or 1.", "clsDFA.LoadDFAFromText"): Exit Sub
                End If
                
                strCurrentLine = Trim$(Mid$(strCurrentLine, i + 1))
                
                If (mblnTextualTapeCellValues = True) Then
                    ' Search for the comma.
                    i = InStr(1, strCurrentLine, ",")
                    
                    If (i > 0) Then _
                        strCurrentLine = Trim$(Mid$(strCurrentLine, i + 1))
                End If
                
            Else
                ' This rule definition parameter isn't optional.
                RaiseEvent ErrorRaised(deInvalidState, "The text isn't a valid DFA file; the current rule is missing the new cell contents value.", "clsDFA.LoadDFAFromText")
                Exit Sub
            End If
        
            ' New State ID (Optional).
            If (Left$(strCurrentLine, 1) = """") Then
                i = InStr(2, strCurrentLine, """")
                
                ' Illegal Rule definition.
                If (i = 0) Then _
                    RaiseEvent ErrorRaised(deInvalidState, "The current rule's ('" & strOriginalLine & "') New State ID property is missing its closing quotation mark.", "clsDFA.LoadDFAFromText"): Exit Sub
                
                clsCurrentState.NewStateID = Trim$(Mid$(strCurrentLine, 2, i - 2))
                
                i = InStr(i, strCurrentLine, ",")

                If (i > 0) Then _
                    strCurrentLine = Trim$(Mid$(strCurrentLine, i + 1))
                
                'If (clsCurrentState.NewStateID <> clsCurrentState.StateID) Then
                '    If (StateIDExists(clsCurrentState.NewStateID) = False) Then
                '        ' New State ID doesn't exist.
                '        RaiseEvent ErrorRaised(deInvalidStateID, "The New State ID, '" & clsCurrentState.NewStateID & "', of the state '" & strOriginalLine & "', doesn't exist!", "clsDFA.LoadDFAFromText")
                '        Exit Sub
                '    End If
                'End If
            
            ElseIf ((Left$(strCurrentLine, 1) = ",") Or (strCurrentLine = "")) Then
                ' Empty State ID ( = Current State ID).
                clsCurrentState.NewStateID = clsCurrentState.StateID
                
                strCurrentLine = Trim$(Mid$(strCurrentLine, 2))
            End If

            ' New Direction (Optional).
            If ((LCase$(Left$(strCurrentLine, 1)) = "l") Or (LCase$(Left$(strCurrentLine, 1)) = "r") Or _
                (LCase$(Left$(strCurrentLine, 1)) = "n") Or (IsNumeric(Left$(strCurrentLine, 1)) = True)) Then
                If (LCase$(Left$(strCurrentLine, 4)) = "left") Then
                    ' One step left.
                    clsCurrentState.NewDirection = -1
                    strCurrentLine = Trim$(Mid$(strCurrentLine, 5))
                ElseIf (LCase$(Left$(strCurrentLine, 5)) = "right") Then
                    ' One step right.
                    clsCurrentState.NewDirection = 1
                    strCurrentLine = Trim$(Mid$(strCurrentLine, 6))
                ElseIf ((LCase$(Left$(strCurrentLine, 7)) = "nowhere") And _
                    (mblnReadHeadCanStayAtPosition = True)) Then
                    ' Move nowhere.
                    clsCurrentState.NewDirection = 0
                    strCurrentLine = Trim$(Mid$(strCurrentLine, 8))
                Else
                    i = InStr(1, strCurrentLine, ",")
                    If (i = 0) Then i = Len(strCurrentLine) + 1
                    
                    If (LCase$(Trim$(Mid$(strCurrentLine, 1, i - 1))) = "l") Then
                        ' One step left.
                        clsCurrentState.NewDirection = -1
                        strCurrentLine = Trim$(Mid$(strCurrentLine, 2))
                    ElseIf (LCase$(Trim$(Mid$(strCurrentLine, 1, i - 1))) = "r") Then
                        ' One step right.
                        clsCurrentState.NewDirection = 1
                        strCurrentLine = Trim$(Mid$(strCurrentLine, 2))
                    ElseIf ((LCase$(Trim$(Mid$(strCurrentLine, 1, i - 1))) = "n") And _
                        (mblnReadHeadCanStayAtPosition = True)) Then
                        ' Move nowhere
                        clsCurrentState.NewDirection = 0
                        strCurrentLine = Trim$(Mid$(strCurrentLine, 2))
                    Else
                        ' Illegal Rule definition.
                        If (IsNumeric(Mid$(strCurrentLine, 1, 1)) = False) Then _
                            RaiseEvent ErrorRaised(deInvalidDFAStateValue, "The current rule's ('" & strOriginalLine & "') New Direction property has an invalid value.", "clsDFA.LoadDFAFromText"): Exit Sub
                        
                        ' New direction is more than one step.
                        intTemp = 0
                        For c = 1 To i - 1
                            If (IsNumeric(Mid$(strCurrentLine, c, 1)) = True) Then
                                intTemp = (intTemp * 10) + CByte(Mid$(strCurrentLine, c, 1))
                            Else
                                Exit For
                            End If
                        Next c
                    
                        If ((LCase$(Trim$(Mid$(strCurrentLine, c, i - c))) = "l") Or _
                            (LCase$(Trim$(Mid$(strCurrentLine, c, i - c))) = "left")) Then
                            ' intTemp steps left.
                            clsCurrentState.NewDirection = -1 * intTemp
                            strCurrentLine = Trim$(Mid$(strCurrentLine, c))
                            If (LCase$(Mid$(strCurrentLine, 1, 4)) = "left") Then
                                strCurrentLine = Trim$(Mid$(strCurrentLine, 5))
                            Else
                                strCurrentLine = Trim$(Mid$(strCurrentLine, 2))
                            End If
                        ElseIf ((LCase$(Trim$(Mid$(strCurrentLine, c, i - c))) = "r") Or _
                            (LCase$(Trim$(Mid$(strCurrentLine, c, i - c))) = "right")) Then
                            ' intTemp steps right.
                            clsCurrentState.NewDirection = 1 * intTemp
                            strCurrentLine = Trim$(Mid$(strCurrentLine, c))
                            If (LCase$(Mid$(strCurrentLine, 1, 5)) = "right") Then
                                strCurrentLine = Trim$(Mid$(strCurrentLine, 6))
                            Else
                                strCurrentLine = Trim$(Mid$(strCurrentLine, 2))
                            End If
                        ElseIf ((LCase$(Trim$(Mid$(strCurrentLine, c, i - c))) = "") And _
                            (intTemp = 0) And (mblnReadHeadCanStayAtPosition = True)) Then
                            ' Move nowhere.
                            clsCurrentState.NewDirection = 0
                            
                            strCurrentLine = Trim$(Mid$(strCurrentLine, c))
                        Else
                            ' Illegal Rule definition.
                            RaiseEvent ErrorRaised(deInvalidDFAStateValue, "The current rule's ('" & strOriginalLine & "') New Direction property has an invalid value.", "clsDFA.LoadDFAFromText")
                            Exit Sub
                        End If
                        
                        ' Illegal Rule definition.
                        If (((Abs(clsCurrentState.NewDirection) > 0) And (mblnReadHeadCanMoveSeveralSteps = False)) Or _
                            ((clsCurrentState.NewDirection = 0) And (mblnReadHeadCanStayAtPosition = False))) Then _
                            RaiseEvent ErrorRaised(deInvalidStateDirection, "The current rule's ('" & strOriginalLine & "') New Direction property has an invalid value.", "clsDFA.LoadDFAFromText"): Exit Sub
                    End If
                End If
                
            ElseIf ((Left$(strCurrentLine, 1) = ",") Or (strCurrentLine = "")) Then
                ' Empty Direction ( = Right).
                clsCurrentState.NewDirection = 1
                
                strCurrentLine = Trim$(Mid$(strCurrentLine, 2))
            End If
            
            If ((Left$(strCurrentLine, 1) = ",") And (mblnAllowAcceptingStates = True)) Then _
                strCurrentLine = Trim$(Mid$(strCurrentLine, 2))
        
            ' Accepting Rule (Optional).
            If ((Left$(strCurrentLine, 1) <> "") And (mblnAllowAcceptingStates = True)) Then
                If (LCase$(Left$(strCurrentLine, 4)) = "true") Then
                        clsCurrentState.AcceptingState = True
                ElseIf (LCase$(Left$(strCurrentLine, 5)) = "false") Then
                    clsCurrentState.AcceptingState = False
                ElseIf ((LCase$(Left$(strCurrentLine, 1)) = "t") Or _
                    (LCase$(Left$(strCurrentLine, 1)) = "f")) Then
                    If (Left$(Trim$(Mid$(strCurrentLine, 2)), 1) = "") Or _
                        (Left$(Trim$(Mid$(strCurrentLine, 2)), 1) = "'") Then
                        If (LCase$(Left$(strCurrentLine, 1)) = "t") Then
                            clsCurrentState.AcceptingState = True
                        Else
                            clsCurrentState.AcceptingState = False
                        End If
                    Else
                        ' Illegal Rule definition.
                        RaiseEvent ErrorRaised(deInvalidDFAStateValue, "The current rule's ('" & strOriginalLine & "') Accepting Rule property has an invalid value.", "clsDFA.LoadDFAFromText")
                        Exit Sub
                    End If
                Else
                    ' Illegal Rule definition.
                    RaiseEvent ErrorRaised(deInvalidDFAStateValue, "The current rule's ('" & strOriginalLine & "') Accepting Rule property has an invalid value.", "clsDFA.LoadDFAFromText")
                    Exit Sub
                End If
                
            ElseIf (mblnAllowAcceptingStates = True) Then
                ' Empty Accepting State ( = False).
                clsCurrentState.AcceptingState = False
                
            ElseIf ((Left$(strCurrentLine, 1) <> "'") And (strCurrentLine <> "")) Then
                ' Illegal Rule definition.
                RaiseEvent ErrorRaised(deInvalidDFAStatement, "The current statement, '" & strCurrentLine & "', isn't a valid rule definition.", "clsDFA.LoadDFAFromText")
                Exit Sub
            End If
            
            ' Add the new rule to the rule list.
            mcolDFAStates.Add clsCurrentState
        
            strPreviousStateID = clsCurrentState.StateID
            
            Set clsCurrentState = New clsDFAState
            clsCurrentState.StateID = ""
            clsCurrentState.CurrentCellContents = ""
            clsCurrentState.NewCellContents = ""
            clsCurrentState.NewStateID = ""
            clsCurrentState.NewDirection = 0
            clsCurrentState.Comment = ""
            clsCurrentState.AcceptingState = False
            clsCurrentState.BreakPoint = False
        End If
    Loop
    
    If (clsCurrentState.StateID <> "") Then
        ' Add the last rule.
        mcolDFAStates.Add clsCurrentState
    End If
    
    If (mcolDFAStates.Count = 0) Then _
        RaiseEvent ErrorRaised(deNoDFAStates, "The DFA rules are missing!", "clsDFA.LoadDFAFromText"): Exit Sub
    
    ' Validate the states.
    ValidateStates (True)

    ' Reset the DFA simulation.
    ResetSimulation
End Sub


'
' Sub SaveTextualDFA
'
' Purpose:
'   Saves a DFA onto a textual DFA file format: the format is different than
'   SaveDFA sub - it's textual, not binary, so it can be edited by hand using a simple text editor.
'
' Parameters:
'   FileName : The file in which the DFA will be saved to.
'
Public Sub SaveTextualDFA(Filename As String)
    Dim intDFAFile As Integer
    Dim strDFAText As String
    
    ' Overwrite the file if it already exists.
    If (Dir$(Filename) <> "") Then _
        Kill Filename
    
    ' Create the DFA file.
    
    intDFAFile = FreeFile
    
    Open Filename For Binary As intDFAFile
    
    ' Retrieve the DFA in its text format.
    strDFAText = SaveDFAIntoText
    
    ' Save the text into the file.
    Put intDFAFile, , strDFAText
    
    Close intDFAFile
End Sub
 
 
'
' Function SaveDFAIntoText
'
' Purpose:
'   Returns the DFA in a text format (so it can be easily edited by hand).
'
' Parameters:
'   JustHeader : If set to True, the function returns only the header of the DFA,
'                without any rules\states.
'
' Returns:
'   The text format of the DFA.
'
Public Function SaveDFAIntoText(Optional JustHeader As Boolean = False) As String
    ' Contains the text format of the DFA.
    Dim strRet As String
    Dim clsCurrentState As clsDFAState
    
    Dim strTemp As String
    
    ' Write the DFA header.
    
    ' Write the DFA's name, description and author.
    strRet = "DFA Name = " & mstrDFAName & vbCrLf
    strTemp = ReplaceString(mstrDFADescription, "\", "\\")
    strTemp = ReplaceString(strTemp, vbCrLf, "\n")
    strTemp = ReplaceString(strTemp, vbTab, "\t")
    strRet = strRet & "DFA Description = " & strTemp & vbCrLf & _
        "DFA Author = " & mstrDFAAuthor
    
    ' Write the advanced options.
    strRet = strRet & vbCrLf & vbCrLf & _
        "Textual Tape Cells = " & mblnTextualTapeCellValues & vbCrLf & _
        "Read Head Can Move Several Steps = " & mblnReadHeadCanMoveSeveralSteps & vbCrLf & _
        "Read Head Can Stay At Same Position = " & mblnReadHeadCanStayAtPosition & vbCrLf & _
        "Allow Accepting Rules = " & mblnAllowAcceptingStates
    
    ' Write the ID of the initial state.
    strRet = strRet & vbCrLf & vbCrLf & _
        "Initial State = " & mstrInitialStateID
    
    strRet = strRet & vbCrLf & vbCrLf & "DFA Code:" & vbCrLf
    
    If (JustHeader = True) Then
        ' Return only the DFA header.
        SaveDFAIntoText = strRet
        Exit Function
    End If
    
    ' Write all of the DFA states.
    
    For Each clsCurrentState In mcolDFAStates
        ' Write the state's comment (if it has a comment).
        If (clsCurrentState.Comment <> "") Then _
            strRet = strRet & vbCrLf & "' " & clsCurrentState.Comment
        ' If the current state is a breakpoint, write "break:".
        If (clsCurrentState.BreakPoint = True) Then
            strRet = strRet & vbCrLf & "Break: "
            ' Write the current state ID.
            strRet = strRet & """" & clsCurrentState.StateID & """, "
        Else
            ' Write the current state ID.
            strRet = strRet & vbCrLf & """" & clsCurrentState.StateID & """, "
        End If
        ' Write the current cell contents.
        If (mblnTextualTapeCellValues) Then
            strRet = strRet & """" & clsCurrentState.CurrentCellContents & """, "
        Else
            strRet = strRet & clsCurrentState.CurrentCellContents & ", "
        End If
        ' Write the new cell contents.
        If (mblnTextualTapeCellValues) Then
            strRet = strRet & """" & clsCurrentState.NewCellContents & """, "
        Else
            strRet = strRet & clsCurrentState.NewCellContents & ", "
        End If
        ' Write the new state ID.
        strRet = strRet & """" & clsCurrentState.NewStateID & """, "
        ' Write the new direction.
        If ((mblnReadHeadCanStayAtPosition = True) And (clsCurrentState.NewDirection = 0)) Then
            strRet = strRet & "Nowhere"
        ElseIf (mblnReadHeadCanMoveSeveralSteps = True) Then
            strRet = strRet & Abs(clsCurrentState.NewDirection) & " " & IIf(clsCurrentState.NewDirection > 0, "Right", "Left")
        Else
            strRet = strRet & IIf(clsCurrentState.NewDirection > 0, "Right", "Left")
        End If
        ' Write if this state is an accepting state (if the "Allow accepting states" option is set).
        If (mblnAllowAcceptingStates) Then _
            strRet = strRet & ", " & clsCurrentState.AcceptingState
    Next clsCurrentState
    
    SaveDFAIntoText = strRet
End Function


'
' Sub NewTape
'
' Purpose:
'   Resets the tape cells.
'
' Note:
'   This sub resizes the tape to a default size, and saves a default value onto the cells.
'
Public Sub NewTape()
    ' Resize the tape to a default value.
    ReDim mstrTapeCells(1 To DEFAULT_TAPE_SIZE)
    
    ' Fill the tape with the default tape cell value.
    FillTapeWithValue (DEFAULT_TAPE_CELL_VALUE)
    
    ' Set the new cell position.
    mintCurrentTapePosition = 1
    
    ReDim mblnTapeCellsBreakpoints(1 To DEFAULT_TAPE_SIZE)
End Sub


'
' Sub LoadTape
'
' Purpose:
'   Loads tape cells from a file.
'
' Parameters:
'   FileName : The file in which the tape cells are located.
'
' Note:
'   This sub overwrites any existing tape cells when loading a new tape from file.
'
Public Sub LoadTape(Filename As String)
    Dim intTapeFile As Integer
    Dim clsCurrentState As clsDFAState
    Dim intTapeCellCount As Integer
    Dim blnTextualCellValues As Boolean
    
    Dim bytTemp As Byte, blnTemp As Boolean, i As Integer

    
    ' The simulation mustn't be running in order to call this sub.
    If (mssSimulationStatus = ssRunning) Then _
        RaiseEvent ErrorRaised(deSubCannotRunWhileSimulatorRunning, "The sub cannot run while the DFA simulator is running", "clsDFA.LoadTape"): Exit Sub
    
    ' File doesn't exist.
    If (Dir$(Filename) = "") Then _
        RaiseEvent ErrorRaised(deFileDoesntExist, "The file '" & Filename & "' doesn't exist", "clsDFA.LoadTape"): Exit Sub
    ' Not a valid tape file.
    If (FileLen(Filename) < 9) Then _
        RaiseEvent ErrorRaised(deInvalidTapeFile, "The file '" & Filename & "' isn't a valid Tape file", "clsDFA.LoadTape"): Exit Sub
    
    ' Open the DFA file.
    
    intTapeFile = FreeFile
    
    Open Filename For Binary As intTapeFile
    
    ' Read the DFA header.
    
    ' The identification string.
    
    Get intTapeFile, , bytTemp
    ' Not a valid tape file.
    If (bytTemp <> Asc("T")) Then _
        RaiseEvent ErrorRaised(deInvalidTapeFile, "The file '" & Filename & "' isn't a valid Tape file", "clsDFA.LoadTape"): Exit Sub
    Get intTapeFile, , bytTemp
    ' Not a valid tape file.
    If (bytTemp <> Asc("A")) Then _
        RaiseEvent ErrorRaised(deInvalidTapeFile, "The file '" & Filename & "' isn't a valid Tape file", "clsDFA.LoadTape"): Exit Sub
    Get intTapeFile, , bytTemp
    ' Not a valid tape file.
    If (bytTemp <> Asc("P")) Then _
        RaiseEvent ErrorRaised(deInvalidTapeFile, "The file '" & Filename & "' isn't a valid Tape file", "clsDFA.LoadTape"): Exit Sub
    
    ' Are the cell values textual?
    Get intTapeFile, , blnTextualCellValues
    
    ' The tape is textual while the DFA doesn't support textual tape cell values.
    If ((mblnTextualTapeCellValues = False) And (blnTextualCellValues = True)) Then _
        RaiseEvent ErrorRaised(deInvalidTapeProperty, "The tape cells are textual, while the current DFA doesn't support textual tape cell values.", "clsDFA.LoadTape"): Exit Sub
    
    ' Read the total no. of tape cells.
    Get intTapeFile, , intTapeCellCount
    
    ' Not a valid tape file.
    If ((intTapeCellCount <= 0) Or (EOF(intTapeFile))) Then _
        RaiseEvent ErrorRaised(deInvalidTapeFile, "The file '" & Filename & "' isn't a valid Tape file", "clsDFA.LoadTape"): Exit Sub
    
    ' Redimension the tape cell array.
    ReDim mstrTapeCells(1 To intTapeCellCount)
    ReDim mblnTapeCellsBreakpoints(1 To intTapeCellCount)
    
    ' Read all of the tape cells.
    
    If (Not blnTextualCellValues) Then
        ' Read the boolean values (zero\one) of the cells.
        
        For i = 1 To intTapeCellCount
            ' Not a valid tape file.
            If (EOF(intTapeFile)) Then _
                RaiseEvent ErrorRaised(deInvalidTapeFile, "The file '" & Filename & "' isn't a valid Tape file", "clsDFA.LoadTape"): Exit Sub
            
            ' Read the current tape cell.
            Get intTapeFile, , bytTemp
            mstrTapeCells(i) = CStr(bytTemp)
        Next i
    Else
        ' Read the textual values of the cells.
        
        For i = 1 To intTapeCellCount
            ' Not a valid tape file.
            If (EOF(intTapeFile)) Then _
                RaiseEvent ErrorRaised(deInvalidTapeFile, "The file '" & Filename & "' isn't a valid Tape file", "clsDFA.LoadTape"): Exit Sub
            
            ' Read the current tape cell.
            mstrTapeCells(i) = ReadString(intTapeFile)
        Next i
    End If
    
    ' Read the indexes of the cells which have breakpoints.
    
    Get intTapeFile, , i
    Do While (Not EOF(intTapeFile))
        ' Not a valid tape file.
        If ((i < 1) Or (i > intTapeCellCount)) Then _
            RaiseEvent ErrorRaised(deInvalidTapeFile, "The file '" & Filename & "' isn't a valid Tape file", "clsDFA.LoadTape"): Exit Sub
        
        mblnTapeCellsBreakpoints(i) = True
        Get intTapeFile, , i
    Loop
    
    Close intTapeFile

    ' Set the new cell position.
    mintCurrentTapePosition = LBound(mstrTapeCells)

    ' Reset the DFA simulation.
    ResetSimulation
End Sub


'
' Sub SaveTape
'
' Purpose:
'   Saves the current tape cells onto a file.
'
' Parameters:
'   FileName : The file in which the tape will be saved.
'
Public Sub SaveTape(Filename As String)
    Dim intTapeFile As Integer
    Dim clsCurrentState As clsDFAState
    
    Dim i As Integer

    ' Overwrite the file if it already exists.
    If (Dir$(Filename) <> "") Then _
        Kill Filename
    
    ' Create the tape file.
    
    intTapeFile = FreeFile
    
    Open Filename For Binary As intTapeFile
    
    ' Write the tape header.
    
    ' Write some sort of identification string.
    Put intTapeFile, , CByte(Asc("T"))
    Put intTapeFile, , CByte(Asc("A"))
    Put intTapeFile, , CByte(Asc("P"))
    
    ' Are the tape values textual?
    Put intTapeFile, , mblnTextualTapeCellValues

    ' Write the total no. of tape cells.
    Put intTapeFile, , CInt(UBound(mstrTapeCells) - LBound(mstrTapeCells) + 1)
    
    ' Write all of the tape cells.
    
    If (Not mblnTextualTapeCellValues) Then
        ' Write the boolean values (zero\one) of the cells.
        For i = LBound(mstrTapeCells) To UBound(mstrTapeCells)
            ' Write the current cell value.
            Put intTapeFile, , CByte(mstrTapeCells(i))
        Next i
    Else
        ' Write the textual values of the cells.
        For i = LBound(mstrTapeCells) To UBound(mstrTapeCells)
            ' Write the current cell value.
            WriteString intTapeFile, mstrTapeCells(i)
        Next i
    End If
    
    ' Write the indexes of the cells which have breakpoints.
    For i = LBound(mblnTapeCellsBreakpoints) To UBound(mblnTapeCellsBreakpoints)
        If (mblnTapeCellsBreakpoints(i) = True) Then
            Put intTapeFile, , i
        End If
    Next i
    
    Close intTapeFile
End Sub


'
' Sub SaveTextualTape
'
' Purpose:
'   Saves a tape onto a textual tape file format: the format is different than
'   SaveTape sub - it's textual, not binary, so it can be edited by hand using a simple text editor.
'
' Parameters:
'   FileName : The file in which the tape will be saved to.
'
Public Sub SaveTextualTape(Filename As String)
    Dim intTapeFile As Integer
    Dim strTapeText As String
    
    ' Overwrite the file if it already exists.
    If (Dir$(Filename) <> "") Then _
        Kill Filename
    
    ' Create the tape file.
    
    intTapeFile = FreeFile
    
    Open Filename For Binary As intTapeFile
    
    ' Retrieve the tape in its text format.
    strTapeText = SaveTapeIntoText
    
    ' Save the text into the file.
    Put intTapeFile, , strTapeText
    
    Close intTapeFile
End Sub


'
' Function SaveTapeIntoText
'
' Purpose:
'   Returns the tape in a text format (so it can be easily edited by hand).
'
' Parameters:
'   NoHeader : If set to True, the function doesn't the header of the tape.
'
' Returns:
'   The text format of the tape.
'
Public Function SaveTapeIntoText(Optional NoHeader As Boolean = False) As String
    ' Contains the text format of the tape.
    Dim strRet As String
    
    Dim i As Integer
    
    If (UBound(mstrTapeCells) - LBound(mstrTapeCells) <= 0) Then _
        SaveTapeIntoText = "": Exit Function
    
    strRet = ""
    
    If (NoHeader = False) Then
        ' Write the tape header.
        strRet = "Textual Tape Cells = " & mblnTextualTapeCellValues & vbCrLf & _
            "Tape Cell Count = " & (UBound(mstrTapeCells) - LBound(mstrTapeCells) + 1) & vbCrLf & vbCrLf & _
            "Tape Cells:" & vbCrLf & vbCrLf
    End If
    
    If (mblnTextualTapeCellValues = False) Then
        For i = LBound(mstrTapeCells) To UBound(mstrTapeCells)
            strRet = strRet & IIf(mblnTapeCellsBreakpoints(i) = True, "Break: ", "") & mstrTapeCells(i)
            
            If (i < UBound(mstrTapeCells)) Then
                strRet = strRet & ", "
                If (i Mod 20 = 0) Then _
                    strRet = strRet & vbCrLf
            End If
        Next i
    Else
        For i = LBound(mstrTapeCells) To UBound(mstrTapeCells)
            strRet = strRet & IIf(mblnTapeCellsBreakpoints(i) = True, "Break: ", "") & """" & mstrTapeCells(i) & """"
            
            If (i < UBound(mstrTapeCells)) Then
                strRet = strRet & ", "
                If (i Mod 10 = 0) Then _
                    strRet = strRet & vbCrLf
            End If
        Next i
    End If

    SaveTapeIntoText = strRet
End Function


'
' Sub LoadTextualTape
'
' Purpose:
'   Loads a Textual tape from a file (the tape file format is different - it's textual,
'   not binary, so it can be edited by hand using a simple text editor).
'
' Parameters:
'   FileName : The file in which the tape is located.
'
' Note:
'   This subs overwrites the previous tape cells while loading the new tape.
'
Public Sub LoadTextualTape(Filename As String)
    Dim intTapeFile As Integer
    Dim strTapeFile As String
    
    ' The simulation mustn't be running in order to call this sub.
    If (mssSimulationStatus = ssRunning) Then _
        RaiseEvent ErrorRaised(deSubCannotRunWhileSimulatorRunning, "The sub cannot run while the DFA simulator is running", "clsDFA.LoadTextualTape"): Exit Sub
    
    ' File doesn't exist.
    If (Dir$(Filename) = "") Then _
        RaiseEvent ErrorRaised(deFileDoesntExist, "The file '" & Filename & "' doesn't exist", "clsDFA.LoadTextualTape"): Exit Sub
    ' Not a valid tape file.
    If (FileLen(Filename) = 0) Then _
        RaiseEvent ErrorRaised(deInvalidDFAFile, "The file '" & Filename & "' isn't a valid tape file", "clsDFA.LoadTextualTape"): Exit Sub
    
    ' Open the textual tape file.
    
    intTapeFile = FreeFile
    
    Open Filename For Input As intTapeFile
        ' Read the contents of the entire file.
        strTapeFile = Input(LOF(intTapeFile), intTapeFile)
    Close intTapeFile
    
    ' Read the contents of the textual tape file.
    LoadTapeFromText (strTapeFile)
End Sub


'
' Sub LoadTapeFromText
'
' Purpose:
'   Loads a tape from a given text segment.
'
' Parameters:
'   Text : The text in which the tape is defined.
'
' Note:
'   This sub overwrites any previous tape cells.
'
Public Sub LoadTapeFromText(ByVal Text As String)
    Dim intCurrentCell As Integer, strCurrentCellValue As String
    Dim blnTextualCells As Boolean
    Dim blnErrorRaised As Boolean
    Dim intCellCount As Integer

    Dim strCurrentLine As String, strOriginalLine As String
    Dim strPreviousStateID As String
    
    Dim blnHasComma As Boolean

    Dim intTemp As Integer, i As Integer, c As Integer
    
    On Error GoTo ErrHandler

    ' The simulation mustn't be running in order to call this sub.
    If (mssSimulationStatus = ssRunning) Then _
        RaiseEvent ErrorRaised(deSubCannotRunWhileSimulatorRunning, "The sub cannot run while the DFA simulator is running", "clsDFA.LoadTapeFromText"): Exit Sub
    
    ' Not a valid tape text.
    If (Trim$(Text) = "") Then _
        RaiseEvent ErrorRaised(deInvalidTapeFile, "The text isn't a valid Tape (it's empty).", "clsDFA.LoadTapeFromText"): Exit Sub

    ' Read the tape header.
    
    Do While (Text <> "")
        strCurrentLine = ""
        
        ' Skip any blank\comment lines.
        Do While (((strCurrentLine = "") Or (Left$(strCurrentLine, 1) = "'")) And (Text <> ""))
            i = InStr(1, Text, vbCrLf)
            If (i = 0) Then i = Len(Text) + 1
            
            strCurrentLine = Trim$(Left$(Text, i - 1))
            
            If (i > Len(Text)) Then
                Text = ""
            Else
                Text = Mid$(Text, i + 2)
            End If
        Loop
        
        strOriginalLine = strCurrentLine
        
        If (LCase$(Left$(strCurrentLine, 18)) = "textual tape cells") Then
            ' Textual Tape Cells property.
            
            strCurrentLine = Trim$(Mid$(strCurrentLine, 19))
            
            If (Left$(strCurrentLine, 1) <> "=") Then _
                RaiseEvent ErrorRaised(deInvalidTapeProperty, "The current tape property, 'Textual Tape Cells', has no value set (either True\T or False\F).", "clsDFA.LoadTapeFromText"): Exit Sub
            
            If ((LCase$(Trim$(Mid$(strCurrentLine, 2))) = "true") Or _
                (LCase$(Trim$(Mid$(strCurrentLine, 2))) = "t")) Then
                ' The tape is textual while the DFA doesn't support textual tape cell values.
                If (mblnTextualTapeCellValues = False) Then _
                    RaiseEvent ErrorRaised(deInvalidTapeProperty, "The tape cells are textual, while the current DFA doesn't support textual tape cell values.", "clsDFA.LoadTapeFromText"): Exit Sub
                blnTextualCells = True
            ElseIf ((LCase$(Trim$(Mid$(strCurrentLine, 2))) = "false") Or _
                (LCase$(Trim$(Mid$(strCurrentLine, 2))) = "f")) Then
                blnTextualCells = False
            Else
                RaiseEvent ErrorRaised(deInvalidTapeProperty, "The current tape property, 'Textual Tape Cells', has an invalid value set - '" & Mid$(strCurrentLine, 2) & "'." & vbCrLf & "It can only be either True\T or False\F.", "clsDFA.LoadTapeFromText")
                Exit Sub
            End If
            
        ElseIf (LCase$(Left$(strCurrentLine, 15)) = "tape cell count") Then
            ' Tape Cell Count property.
            
            strCurrentLine = Trim$(Mid$(strCurrentLine, 16))
            
            If (Left$(strCurrentLine, 1) <> "=") Then _
                RaiseEvent ErrorRaised(deInvalidTapeProperty, "The current tape property, 'Textual Tape Cells', has no value set (either True\T or False\F).", "clsDFA.LoadTapeFromText"): Exit Sub
            
            blnErrorRaised = False
            intCellCount = CInt(Trim$(Mid$(strCurrentLine, 2)))
            
            If ((blnErrorRaised = True) Or (intCellCount <= 0)) Then _
                RaiseEvent ErrorRaised(deInvalidTapeProperty, "The current tape property, 'Tape Cell Count', has an invalid value set - '" & Mid$(strCurrentLine, 2) & "'.", "clsDFA.LoadTapeFromText"): Exit Sub
                
        ElseIf (LCase$(Left$(strCurrentLine, 10)) = "tape cells") Then
            ' Marks the beginning of the tape cells (and the end of the tape header).
            
            If ((Trim$(Mid$(strCurrentLine, 11)) <> ":") And (Trim$(Mid$(strCurrentLine, 10)) <> "")) Then _
                RaiseEvent ErrorRaised(deInvalidTapeStatement, "The current statement, '" & strCurrentLine & "' is invalid. Did you mean to write 'Tape Cells:'?", "clsDFA.LoadTapeFromText"): Exit Sub
            
            strCurrentLine = Trim$(Mid$(strCurrentLine, 9))
            
            Exit Do
            
        ElseIf (Text <> "") Then
            ' Invalid tape property.
            
            RaiseEvent ErrorRaised(deInvalidTapeStatement, "The current tape property, '" & strCurrentLine & "', is an invalid tape property.", "clsDFA.LoadTapeFromText")
            Exit Sub
        End If
    Loop
    
    
    ' Reset the tape cells.
    
    ReDim mstrTapeCells(1 To intCellCount)
    ReDim mblnTapeCellsBreakpoints(1 To intCellCount)
    mintCurrentTapePosition = 1
    
    
    ' Read the tape cells.
    
    intCurrentCell = 0
    blnHasComma = False
    
    Do While (Text <> "")
        strCurrentLine = ""
        
        ' Skip any blank\comment lines.
        Do While (((strCurrentLine = "") Or (Left$(strCurrentLine, 1) = "'")) And (Text <> ""))
            i = InStr(1, Text, vbCrLf)
            If (i = 0) Then i = Len(Text) + 1
            
            strCurrentLine = Trim$(Left$(Text, i - 1))
            
            If (i > Len(Text)) Then
                Text = ""
            Else
                Text = Mid$(Text, i + 2)
            End If
        Loop
        
        strOriginalLine = strCurrentLine
        
        ' Parse the current line.
        Do
            If (LCase$(Left$(strCurrentLine, 6)) = "break:") Then
                ' The current tape cell is a breakpoint.
                
                blnHasComma = False
                
                If (intCurrentCell + 1 > intCellCount) Then _
                    RaiseEvent ErrorRaised(deInvalidTapeStatement, "There are more tape cells than stated in the 'Tape Cell Count' property (" & intCellCount & ").", "clsDFA.LoadTapeFromText"): Exit Sub
                
                mblnTapeCellsBreakpoints(intCurrentCell + 1) = True
                strCurrentLine = Trim$(Mid$(strCurrentLine, 7))
            End If
            
            If ((Left$(strCurrentLine, 1) = """") And (blnTextualCells = True)) Then
                ' A textual cell value.
                intCurrentCell = intCurrentCell + 1
                
                blnHasComma = False
                
                If (intCurrentCell > intCellCount) Then _
                    RaiseEvent ErrorRaised(deInvalidTapeStatement, "There are more tape cells than stated in the 'Tape Cell Count' property (" & intCellCount & ").", "clsDFA.LoadTapeFromText"): Exit Sub
                
                ' Find the closing quotation mark.
                i = InStr(2, strCurrentLine, """")
                
                If (i = 0) Then _
                    RaiseEvent ErrorRaised(deInvalidTapeStatement, "The current tape cell in the line '" & strOriginalLine & "' is missing its closing quotation mark.", "clsDFA.LoadTapeFromText"): Exit Sub
                
                mstrTapeCells(intCurrentCell) = Mid$(strCurrentLine, 2, i - 2)
                strCurrentLine = Trim$(Mid$(strCurrentLine, i + 1))
            
            ElseIf (((Left$(strCurrentLine, 1) = "0") Or (Left$(strCurrentLine, 1) = "1")) _
                And (blnTextualCells = False)) Then
                ' A non-textual cell value.
                
                blnHasComma = False
                
                intCurrentCell = intCurrentCell + 1
                
                If (intCurrentCell > intCellCount) Then _
                    RaiseEvent ErrorRaised(deInvalidTapeStatement, "There are more tape cells than stated in the 'Tape Cell Count' property (" & intCellCount & ").", "clsDFA.LoadTapeFromText"): Exit Sub
                
                mstrTapeCells(intCurrentCell) = Left$(strCurrentLine, 1)
                strCurrentLine = Trim$(Mid$(strCurrentLine, 2))
            ElseIf (Left$(strCurrentLine, 1) = ",") Then
                ' The current cell value was omitted.
                
                blnHasComma = False
                
                intCurrentCell = intCurrentCell + 1
                
                If (intCurrentCell > intCellCount) Then _
                    RaiseEvent ErrorRaised(deInvalidTapeStatement, "There are more tape cells than stated in the 'Tape Cell Count' property (" & intCellCount & ").", "clsDFA.LoadTapeFromText"): Exit Sub
                
                
                If (intCurrentCell = 1) Then
                    mstrTapeCells(intCurrentCell) = DEFAULT_TAPE_CELL_VALUE
                Else
                    mstrTapeCells(intCurrentCell) = mstrTapeCells(intCurrentCell - 1)
                End If
                
            ElseIf (strCurrentLine <> "") Then
                ' Invalid cell value.
                RaiseEvent ErrorRaised(deInvalidTapeStatement, "The current tape cell value in the line '" & strOriginalLine & "' is invalid.", "clsDFA.LoadTapeFromText")
                Exit Sub
            End If
            
            ' Search for the comma.
            If ((Left$(strCurrentLine, 1) <> ",") And (Left$(strCurrentLine, 1) <> "'") And _
                (strCurrentLine <> "")) Then _
                RaiseEvent ErrorRaised(deInvalidTapeStatement, "The current tape cell value in the line '" & strOriginalLine & "' is invalid.", "clsDFA.LoadTapeFromText"): Exit Sub
            
            If (Left$(strCurrentLine, 1) = "'") Then
                strCurrentLine = ""
            ElseIf (Left$(strCurrentLine, 1) = ",") Then _
                blnHasComma = True
            End If
                
            strCurrentLine = Trim$(Mid$(strCurrentLine, 2))
        Loop While (strCurrentLine <> "")
    Loop
    
    If (blnHasComma = True) Then
        ' Missing additional cell value.
        RaiseEvent ErrorRaised(deInvalidTapeStatement, "A tape cell value is missing (a comma was written at the end of a line, but no cell value was provided afterwards).", "clsDFA.LoadTapeFromText")
        Exit Sub
    End If
    
    If (intCurrentCell < intCellCount) Then
        ' Fill any cells not filled by the user.
        For i = intCurrentCell + 1 To intCellCount
            mstrTapeCells(i) = DEFAULT_TAPE_CELL_VALUE
        Next i
    End If
    
    
    Exit Sub

ErrHandler:
    blnErrorRaised = True
    
    Resume Next
End Sub


'
' Sub FillTapeRandomly
'
' Purpose:
'   Fills the tape cells with randomly chosen values (either zero or one).
'
Public Sub FillTapeRandomly()
    Dim i As Integer
    
    ' The simulation mustn't be running in order to call this sub.
    If (mssSimulationStatus = ssRunning) Then _
        RaiseEvent ErrorRaised(deSubCannotRunWhileSimulatorRunning, "The sub cannot run while the DFA simulator is running", "clsDFA.FillTapeRandomly"): Exit Sub
    
    Randomize
    
    For i = LBound(mstrTapeCells) To UBound(mstrTapeCells)
        mstrTapeCells(i) = CStr(Int(2 * Rnd))
    Next i
End Sub


'
' Sub FillTapeWithValue
'
' Purpose:
'   Fills the tape cells with a value specified by the user.
'
' Paremeters:
'   CellValue : The value which will be used to fill the tape cells with.
'   (Optional - default value: "0")
'
Public Sub FillTapeWithValue(Optional CellValue As String = DEFAULT_TAPE_CELL_VALUE)
    Dim i As Integer
    
    ' The simulation mustn't be running in order to call this sub.
    If (mssSimulationStatus = ssRunning) Then _
        RaiseEvent ErrorRaised(deSubCannotRunWhileSimulatorRunning, "The sub cannot run while the DFA simulator is running", "clsDFA.FillTapeWithValue"): Exit Sub
    
    ' Invalid cell value (since textual cell values option is not set).
    If ((Not mblnTextualTapeCellValues) And (CellValue <> "0") And (CellValue <> "1")) Then _
        RaiseEvent ErrorRaised(deInvalidCellValue, "Invalid cell value entered: " & CellValue, "clsDFA.FillTapeWithValue"): Exit Sub
    
    For i = LBound(mstrTapeCells) To UBound(mstrTapeCells)
        mstrTapeCells(i) = CellValue
    Next i
End Sub


'
' Private Subs
'


'
' Sub RunSimulation
'
' Purpose:
'   Runs the DFA simulation.
'
Private Sub RunSimulation()
    ' The current state being executed.
    Dim clsCurrentState As clsDFAState
    ' Used for determining if we already paused the simulation for the same breakpoint.
    Static blnReachedBreakPoint As Boolean
    ' Indicates whether the sub is already running.
    Static blnAlreadyRunning As Boolean
    ' Passed by reference to the ReachedState event.
    Dim colExitSub As New Collection
    
    Dim i As Integer
    
    
    On Error Resume Next
    
    ' The simulation is already running.
    If ((mssSimulationStatus = ssRunning) Or (blnAlreadyRunning = True)) Then _
        mssSimulationStatus = ssRunning: Exit Sub
    
    blnAlreadyRunning = True

    colExitSub.Add False
    
    If (mssSimulationStatus = ssComplete) Then
        ' The simulation is complete\stopped - reset its properties.
        mstrCurrentStateID = mstrInitialStateID
        mintCurrentTapePosition = LBound(mstrTapeCells)
    End If
        
    
    ' No DFA states are available.
    If (mcolDFAStates Is Nothing) Then _
        RaiseEvent ErrorRaised(deNoDFAStates, "The DFA has no states", "clsDFA.RunSimulation"): Exit Sub
    If (mcolDFAStates.Count = 0) Then _
        RaiseEvent ErrorRaised(deNoDFAStates, "The DFA has no states", "clsDFA.RunSimulation"): Exit Sub
    
    If (mssSimulationStatus = ssStopped) Then
        ' Start the simulation from the beginning.
        mstrCurrentStateID = mstrInitialStateID
        mintCurrentTapePosition = LBound(mstrTapeCells)
    
        ' We don't have an initial state.
        If (mstrInitialStateID = "") Then _
            RaiseEvent ErrorRaised(deNoInitialDFAState, "The DFA has no initial state", "clsDFA.RunSimulation"): Exit Sub
    End If
    
    ' We don't have the current state ready.
    If (mstrCurrentStateID = "") Then _
        RaiseEvent ErrorRaised(deNoCurrentDFAState, "The DFA has no current state", "clsDFA.RunSimulation"): Exit Sub
    ' The read head isn't pointing to a valid cell in the tape.
    If ((mintCurrentTapePosition < LBound(mstrTapeCells)) Or (mintCurrentTapePosition > UBound(mstrTapeCells))) Then _
        RaiseEvent ErrorRaised(deInvalidReadHeadPosition, "The DFA's read head is pointing at an invalid position in the tape", "clsDFA.RunSimulation"): Exit Sub
    
    mssSimulationStatus = ssRunning
    
    ' Run the DFA simulation.
    
    Do
        ' Find the state matching the current cell value.
        Set clsCurrentState = GetDFAState(mstrCurrentStateID, mstrTapeCells(mintCurrentTapePosition))
        
        ' State not found.
        If (clsCurrentState Is Nothing) Then _
            RaiseEvent ErrorRaised(deDFAStateNotFound, "The current DFA state wasn't found (Current State ID: '" & mstrCurrentStateID & "'; Current Cell Contents: " & mstrTapeCells(mintCurrentTapePosition) & ")", "clsDFA.RunSimulation"): Exit Sub
        
        ' The current state is also a breakpoint.
        If ((clsCurrentState.BreakPoint = True) And (blnReachedBreakPoint = False)) Then
            mssSimulationStatus = ssPaused
            blnReachedBreakPoint = True
            RaiseEvent ReachedBreakpoint(1)
            blnAlreadyRunning = False
            Exit Sub
        End If
        
        ' Reset the breakpoint flag.
        blnReachedBreakPoint = False
        
        ' If we've reached this far, this means we've found a matching state.
        
        ' Call the ReachedState event before executing the current state.
        RaiseEvent ReachedState(clsCurrentState, colExitSub)
        
        ' Loop until the user says it's OK to continue and finish executing the current rule.
        Do While (colExitSub(1) = True)
            DoEvents
        Loop
            
        ' Execute the current state.
        
        ' Set the current tape cell's new value.
        mstrTapeCells(mintCurrentTapePosition) = clsCurrentState.NewCellContents
        ' Move to the new state.
        mstrCurrentStateID = clsCurrentState.NewStateID
        
        ' Move left\right.
        
        ' Invalid direction.
        If (((clsCurrentState.NewDirection = 0) And (Not mblnReadHeadCanStayAtPosition)) Or _
            ((Abs(clsCurrentState.NewDirection) <> 1) And (Not mblnReadHeadCanMoveSeveralSteps))) Then _
            RaiseEvent ErrorRaised(deInvalidStateDirection, "The new direction of the current DFA state is invalid", "clsDFA.RunSimulation"): Exit Sub
        
        mintCurrentTapePosition = mintCurrentTapePosition + clsCurrentState.NewDirection
        
        ' See if we need to stop the simulation.
        If ((mintCurrentTapePosition < LBound(mstrTapeCells)) Or _
            ((clsCurrentState.AcceptingState = True)) And (mblnAllowAcceptingStates)) Then
            blnAlreadyRunning = False
            mssSimulationStatus = ssComplete
            RaiseEvent SimulationComplete(clsCurrentState)
            Exit Do
        End If
        
        ' The current tape cell is a breakpoint.
        If (mblnTapeCellsBreakpoints(mintCurrentTapePosition) = True) Then
            blnAlreadyRunning = False
            mssSimulationStatus = ssPaused
            RaiseEvent ReachedBreakpoint(2)
            Exit Sub
        End If
        
        ' The user chose to pause\stop the simulation.
        If (mssSimulationStatus <> ssRunning) Then _
            Exit Do
    Loop
    
    blnAlreadyRunning = False
End Sub


'
' Sub ResetSimulation
'
' Purpose:
'   Resets the DFA simulation (so when Run is called afterwards, the simulation starts
'   running from beginning). However, this sub doesn't reset the tape (the values remain the same).
'
Private Sub ResetSimulation()
    ' Reset the DFA simulation.
    
    mssSimulationStatus = ssStopped
    mstrCurrentStateID = mstrInitialStateID
    mintCurrentTapePosition = LBound(mstrTapeCells)
End Sub


Private Sub Class_Initialize()
    mssSimulationStatus = ssStopped
    
    ' Initialize the tape.
    NewTape
    
    Set mcolDFAStates = New Collection
End Sub


'
' Function ReadString
'
' Purpose:
'   Reads a null-terminated string from a file.
'
' Parameters:
'   FileNumber : The file no. of the file from which the sub reads the string.
'
' Returns:
'   The string read.
'
Private Function ReadString(FileNumber As Integer) As String
    Dim strOutputString As String
    Dim b As Byte
    
    strOutputString = ""
    
    Get FileNumber, , b
    
    Do While ((Not EOF(FileNumber)) And (b <> 0))
        strOutputString = strOutputString & Chr(b)
        Get FileNumber, , b
    Loop
    
    ReadString = strOutputString
End Function


'
' Sub WriteString
'
' Purpose:
'   Write a null-terminated string onto a file.
'
' Parameters:
'   FileNumber : The file no. of the file onto which the sub writes the string.
'   OutputString : The string that is written onto the file.
'
Private Sub WriteString(FileNumber As Integer, OutputString As String)
    Dim i As Integer
    
    For i = 1 To Len(OutputString)
        Put FileNumber, , CByte(Asc(Mid$(OutputString, i, 1)))
    Next i
    
    ' Write the terminating character (zero).
    Put FileNumber, , CByte(0)
End Sub


'
' Function ReplaceString
'
' Purpose:
'   A Substitute for the Replace function in VB6 (for backward-compatibility with VB5).
'
Private Function ReplaceString(ByVal SearchIn As String, SearchFor As String, ReplaceWith As String, Optional Start As Integer = 1, Optional CompareMethod As VbCompareMethod = vbTextCompare) As String
    Dim i As Integer, c As Integer
    
    i = IIf((Start >= 1) And (Start < Len(SearchIn)), Start, 1)
    
    Do
        c = InStr(i, SearchIn, SearchFor, CompareMethod)
        
        If (c > 1) Then
            SearchIn = Left$(SearchIn, c - 1) & ReplaceWith & Mid$(SearchIn, c + Len(SearchFor))
        ElseIf (c = 1) Then
            SearchIn = ReplaceWith & Mid$(SearchIn, c + Len(SearchFor))
        End If
        
        i = c + Len(ReplaceWith)
    Loop While (c > 0)
    
    ReplaceString = SearchIn
End Function
